from .base_map import base_map
from .batched_string_grid import batched_string_grid
from .caves_cont import caves_cont
from .labmaze_grid import labmaze_grid
from .movingai import movingai
from .random_grid import random_grid
from .string_grid import string_grid

__all__ = [
    "base_map",
    "random_grid",
    "string_grid",
    "batched_string_grid",
    "labmaze_grid",
    "movingai",
    "caves_cont",
]



import jax.numpy as jnp

from camar.maps import base_map


class dummy_map(base_map):
    def __init__(self, **kwargs):
        self.kwargs = kwargs

        self.num_agents = kwargs.get("num_agents", 1)
        self.num_landmarks = kwargs.get("num_landmarks", 1)
        self.obstacle_size = kwargs.get("obstacle_size", 0.2)

    @property
    def landmark_rad(self):
        return self.kwargs.get("landmark_rad", 0.1)

    @property
    def agent_rad(self):
        return self.kwargs.get("agent_rad", 0.1)

    @property
    def goal_rad(self):
        return self.kwargs.get("goal_rad", 0.1)

    @property
    def height(self):
        return self.kwargs.get("height", 10.0)

    @property
    def width(self):
        return self.kwargs.get("width", 10.0)

    def reset(self, key):
        landmark_pos = self.kwargs.get("landmark_pos", jnp.array([0.0, 0.0]))
        agent_pos = self.kwargs.get("agent_pos", jnp.array([1.0, 1.0]))
        goal_pos = self.kwargs.get("goal_pos", jnp.array([2.0, 2.0]))
        return (
            key,
            landmark_pos,
            agent_pos,
            goal_pos,
        )

    def update_goals(self, keys, goal_pos, to_update):
        return keys, goal_pos



import jax

from camar import camar_v0
from camar.maps import labmaze_grid, random_grid, string_grid
from camar.wrappers import OptimisticResetVecEnvWrapper


class TestReadmeExamples:
    def test_quickstart_single_env(self):
        key = jax.random.key(0)
        key, key_r, key_a, key_s = jax.random.split(key, 4)

        env = camar_v0()
        assert env is not None
         
        reset_fn = jax.jit(env.reset)
        step_fn = jax.jit(env.step)

        obs, state = reset_fn(key_r)
        assert obs is not None
        assert state is not None

        actions = env.action_spaces.sample(key_a)
        assert actions.shape == (env.num_agents, env.action_size)

        obs_next, state_next, reward, done, info = step_fn(key_s, state, actions)
        assert obs_next is not None
        assert state_next is not None
        assert reward is not None
        assert done is not None
        assert isinstance(info, dict)

    def test_quickstart_vectorized_env(self):
        key = jax.random.key(0)
        key, key_r, key_a, key_s = jax.random.split(key, 4)

        num_envs = 10  # Reduced from 1000 for faster testing

        env = camar_v0()

        action_sampler = jax.jit(
            jax.vmap(
                env.action_spaces.sample,
                in_axes=[
                    0,
                ],
            )
        )
        env_reset_fn = jax.jit(
            jax.vmap(
                env.reset,
                in_axes=[
                    0,
                ],
            )
        )
        env_step_fn = jax.jit(
            jax.vmap(
                env.step,
                in_axes=[
                    0,
                    0,
                    0,
                ],
            )
        )

        keys_r_v = jax.random.split(key_r, num_envs)
        keys_a_v = jax.random.split(key_a, num_envs)
        keys_s_v = jax.random.split(key_s, num_envs)

        obs, state = env_reset_fn(keys_r_v)
        assert obs.shape == (num_envs, env.num_agents, env.observation_size)

        actions = action_sampler(keys_a_v)
        assert actions.shape == (num_envs, env.num_agents, env.action_size)

        obs_next, state_next, reward, done, info = env_step_fn(keys_s_v, state, actions)
        assert obs_next.shape == (num_envs, env.num_agents, env.observation_size)
        assert reward.shape == (num_envs, env.num_agents, 1)
        assert done.shape == (num_envs,)
        assert isinstance(info, dict)

    def test_wrappers_example(self):
        num_envs = 10  # Reduced from 1000 for faster testing
        env = OptimisticResetVecEnvWrapper(
            env=camar_v0(),
            num_envs=num_envs,
            reset_ratio=2,  # reduced from 20 for faster testing
        )
        assert env is not None

        key = jax.random.key(0)
        key_reset, key_step, key_action = jax.random.split(key, 3)

        obs, state = env.reset(key_reset)
        assert obs.shape[0] == num_envs

        key_actions = jax.random.split(key_action, num_envs)
        actions = jax.vmap(env.action_spaces.sample)(key_actions)
        assert actions.shape == (num_envs, env.num_agents, env.action_size)

        obs_next, state_next, reward, done, info = env.step(key_step, state, actions)
        assert obs_next.shape[0] == num_envs
        assert reward.shape[0] == num_envs
        assert done.shape[0] == num_envs

    def test_maps_example_creation(self):
        map_str_readme = """
        .....#.....
        .....#.....
        ...........
        .....#.....
        .....#.....
        #.####.....
        .....###.##
        .....#.....
        .....#.....
        ...........
        .....#.....
        """

        string_grid_map = string_grid(map_str=map_str_readme, num_agents=8)
        random_grid_map_custom = random_grid(num_agents=4, num_rows=10, num_cols=10)

        env1 = camar_v0(string_grid_map)
        env2 = camar_v0(random_grid_map_custom)

        assert isinstance(env1.map_generator, string_grid)
        assert isinstance(env2.map_generator, random_grid)
        assert env1.num_agents == 8
        assert env2.num_agents == 4

        env1_str = camar_v0("string_grid", map_str=map_str_readme, num_agents=8)
        env2_str = camar_v0("random_grid", num_agents=4, num_rows=10, num_cols=10)

        assert isinstance(env1_str.map_generator, string_grid)
        assert isinstance(env2_str.map_generator, random_grid)
        assert env1_str.num_agents == 8
        assert env2_str.num_agents == 4

        # labmaze is not supported by python=3.13
        try:
            labmaze_map = labmaze_grid(
                num_maps=2, num_agents=3, height=7, width=7
            )  # Reduced for testing
            env3 = camar_v0(labmaze_map)

            assert isinstance(env3.map_generator, labmaze_grid)
            assert env3.num_agents == 3

            env3_str = camar_v0(
                "labmaze_grid", num_maps=2, num_agents=3, height=7, width=7
            )

            assert isinstance(env3_str.map_generator, labmaze_grid)
            assert env3_str.num_agents == 3

        except ModuleNotFoundError:
            pass



import jax
import jax.numpy as jnp

from camar.maps import string_grid

MAP_STR = """
.....#.....
.....#.....
...........
.....#.....
.....#.....
"""


class TestStringGrid:
    def test_map_creation(self):
        map_gen = string_grid(map_str=MAP_STR, num_agents=2)
        assert map_gen is not None
        assert map_gen.num_agents == 2
        assert map_gen.num_landmarks == 4 + 11 * 2 + 5 * 2 + 4

    def test_map_reset(self):
        map_gen = string_grid(map_str=MAP_STR, num_agents=2)
        key = jax.random.key(0)
        key_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert key_g is not None
        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape[0] == map_gen.num_landmarks
        assert landmark_pos.shape[1] == 2
        assert agent_pos.shape == (map_gen.num_agents, 2)
        assert goal_pos.shape == (map_gen.num_agents, 2)

    def test_map_creation_with_specific_agent_goal_pos(self):
        map_gen_no_border = string_grid(
            map_str=MAP_STR,
            num_agents=1,
            agent_idx=jnp.array([[0, 0]]),  # Top-left corner
            goal_idx=jnp.array([[2, 5]]),  # A middle free spot
            add_border=False,
            remove_border=False,
        )
        key = jax.random.key(1)
        _, _, agent_pos, goal_pos = map_gen_no_border.reset(key)

        assert agent_pos is not None
        assert goal_pos is not None
        assert agent_pos.shape == (1, 2)
        assert goal_pos.shape == (1, 2)




import torch
from tensordict import TensorDict, TensorDictBase
from torchrl.data.tensor_specs import Bounded, Categorical, Composite, Unbounded
from torchrl.envs.common import _EnvWrapper
from torchrl.envs.libs.jax_utils import (
    _extract_spec,
    _ndarray_to_tensor,
    _object_to_tensordict,
    _tensor_to_ndarray,
    _tree_flatten,
)
from torchrl.envs.utils import MarlGroupMapType, _classproperty, check_marl_grouping
from camar import camar_v0


class CamarWrapper(_EnvWrapper):
    _jax = None

    @_classproperty
    def jax(cls):
        if cls._jax is not None:
            return cls._jax

        import jax

        cls._jax = jax
        return jax

    def __init__(self, seed, device, batch_size, env = None, **kwargs):
        if env is not None:
            kwargs["env"] = env

        if isinstance(batch_size, int):
            batch_size = [batch_size]

        kwargs["device"] = device
        kwargs["batch_size"] = batch_size
        kwargs["seed"] = seed

        super().__init__(**kwargs)

        self.set_seed(seed)

    def _check_kwargs(self, kwargs: dict):
        if "env" not in kwargs:
            raise TypeError("Could not find environment key 'env' in kwargs.")

    def _build_env(
        self,
        env,
        **kwargs,
    ):
        return env

    def _make_state_spec(self, env):
        jax = self.jax

        key = jax.random.PRNGKey(0)
        state = env.reset(key)
        state_dict = _object_to_tensordict(state, self.device, batch_size=())
        state_spec = _extract_spec(state_dict).expand(self.batch_size)
        return state_spec

    def _make_specs(self, env) -> None:  # noqa: F821
        agent_names = [f"agent_{agent_idx}" for agent_idx in range(env.num_agents)]
        self.group_map = MarlGroupMapType.ALL_IN_ONE_GROUP.get_group_map(agent_names)

        # just to be confident
        check_marl_grouping(self.group_map, agent_names)
        assert len(self.group_map.keys()) == 1
        assert "agents" in self.group_map.keys()

        action = Bounded(
            low=-1,
            high=1,
            shape=(*self.batch_size, env.num_agents, env.action_size),
            device=self.device,
        )
        agents_action = Composite(
            action=action,
            shape=(*self.batch_size, env.num_agents),
            device=self.device,
        )
        self.action_spec = Composite(
            agents=agents_action,
            shape=self.batch_size,
            device=self.device,
        )

        reward = Bounded(
            low=-0.5,
            high=1,
            shape=(*self.batch_size, env.num_agents, 1),
            device=self.device,
        )
        agents_reward = Composite(
            reward=reward,
            shape=(*self.batch_size, env.num_agents),
            device=self.device,
        )
        self.reward_spec = Composite(
            agents=agents_reward,
            shape=self.batch_size,
            device=self.device,
        )

        observation = Unbounded(
            shape=(*self.batch_size, env.num_agents, env.observation_size),
            device=self.device,
        )
        agents_observation = Composite(
            observation=observation,
            shape=(*self.batch_size, env.num_agents),
            device=self.device,
        )
        self.observation_spec = Composite(
            agents=agents_observation,
            shape=self.batch_size,
            device=self.device,
        )

        self.done_spec = Categorical(
            n=2,
            shape=(*self.batch_size, 1),
            dtype=torch.bool,
            device=self.device,
        )

    # def _make_state_example(self):
    #     jax = self.jax

    #     key = jax.random.PRNGKey(0)
    #     keys = jax.random.split(key, self.batch_size.numel())
    #     state, obs, done = self._jit_vmap_env_reset(jax.numpy.stack(keys))
    #     # state = _tree_reshape(state, self.batch_size)
    #     return state

    def _init_env(self):
        jax = self.jax
        self._key = None
        self._jit_vmap_env_reset = jax.jit(jax.vmap(self._env.reset))
        self._jit_vmap_env_step = jax.jit(jax.vmap(self._env.step))
        self._jit_vmap_env_get_obs = jax.jit(jax.vmap(self._env.get_obs))

        self._jit_partial_reset = jax.jit(self._partial_reset)
        self._state = None
        # self._state_example = self._make_state_example()

    def _set_seed(self, seed: int):
        jax = self.jax
        if seed is None:
            raise Exception("CAMAR requires an integer seed.")
        self._key = jax.random.PRNGKey(seed)

    def _partial_reset(self, keys, state, envs_to_reset):
        obs_r, state_r = self._jit_vmap_env_reset(keys)

        obs_old = self._jit_vmap_env_get_obs(state)

        state_old = state

        state = self.jax.tree.map(
            lambda x, y: self.jax.numpy.where(self.jax.numpy.expand_dims(envs_to_reset, range(1, x.ndim)), x, y),
            state_r,
            state_old
        )

        obs = self.jax.numpy.where(envs_to_reset[:, None, None], obs_r, obs_old)

        return obs, state

    def _reset(self, tensordict: TensorDictBase = None, **kwargs) -> TensorDictBase:
        jax = self.jax

        # generate random keys
        self._key, *keys = jax.random.split(self._key, 1 + self.numel())
        keys = jax.numpy.stack(keys)

        if tensordict is not None and "_reset" in tensordict.keys():
            _reset = tensordict.get("_reset")
            envs_to_reset = _reset.squeeze(-1)

            if envs_to_reset.all():
                # reset all
                obs, self._state = self._jit_vmap_env_reset(keys)
            else:
                envs_to_reset = _tensor_to_ndarray(envs_to_reset)

                obs, self._state = self._jit_partial_reset(keys, self._state, envs_to_reset)
        else:
            # call env reset with jit and vmap
            obs, self._state = self._jit_vmap_env_reset(keys)

        tensordict_agents = TensorDict(
            source={
                "observation": _ndarray_to_tensor(obs),
            },
            batch_size=(*self.batch_size, self._env.num_agents),
            device=self.device,
        )

        done = self._state.on_goal.all(axis=-1)
        done = _ndarray_to_tensor(done)

        tensordict_out = TensorDict(
            source={
                "agents": tensordict_agents,
                "done": done,
                "terminated": done.clone(),
            },
            batch_size=self.batch_size,
            device=self.device,
        )
        return tensordict_out

    def _step(self, tensordict: TensorDictBase):
        jax = self.jax

        # convert tensors to ndarrays

        # state = _tensordict_to_object(tensordict.get("state"), self._state_example)
        action = _tensor_to_ndarray(tensordict.get(("agents", "action")))

        # flatten batch size
        # state = _tree_flatten(state, self.batch_size)
        action = _tree_flatten(action, self.batch_size)

        # call env step with jit and vmap
        self._key, *keys_s = jax.random.split(self._key, 1 + self.numel())

        obs, self._state, reward, done, info = self._jit_vmap_env_step(
            jax.numpy.stack(keys_s), self._state, action
        )

        tensordict_agents = TensorDict(
            source={
                "observation": _ndarray_to_tensor(obs),
                "reward": _ndarray_to_tensor(reward),
                "on_goal": _ndarray_to_tensor(self._state.on_goal).view(*self.batch_size, -1, 1),
            },
            batch_size=(*self.batch_size, self._env.num_agents),
            device=self.device,
        )

        time_to_reach_goal = _ndarray_to_tensor(self._state.time_to_reach_goal)
        coordination =  1 - _ndarray_to_tensor(self._state.num_collisions / self._state.step[:, None]).mean(-1) # mean for agents

        flowtime = time_to_reach_goal.sum(dim=-1)
        makespan, _ = time_to_reach_goal.max(dim=-1)

        done = _ndarray_to_tensor(done)

        tensordict_out = TensorDict(
            source={
                "agents": tensordict_agents,
                "done": done,
                "terminated": done.clone(),
                "flowtime": flowtime,
                "makespan": makespan,
                "coordination": coordination,
            },
            batch_size=self.batch_size,
            device=self.device,
        )
        return tensordict_out


class CamarEnv(CamarWrapper):

    def __init__(
        self,
        map_generator,
        num_envs,
        seed,
        max_steps = 1000,
        window = 0.8,
        placeholder = 0.0,
        frameskip = 2,
        max_obs = None,
        dt = 0.01,
        damping = 0.25,
        contact_force = 500,
        contact_margin = 0.001,
        **map_kwargs,
    ):
        batch_size = [num_envs]

        super().__init__(
            batch_size=batch_size,
            map_generator=map_generator,
            num_envs=num_envs,
            max_steps=max_steps,
            seed=seed,
            window=window,
            placeholder=placeholder,
            frameskip=frameskip,
            max_obs=max_obs,
            dt=dt,
            damping=damping,
            contact_force=contact_force,
            contact_margin=contact_margin,
            **map_kwargs,
        )

    def _check_kwargs(self, kwargs: dict):
        if "map_generator" not in kwargs:
            raise TypeError("Cannot find the environment key 'map_generator' in kwargs.")
        if "num_envs" not in kwargs:
            raise TypeError("Cannot find the environment key 'num_envs' in kwargs.")

    def _build_env(
        self,
        map_generator,
        num_envs,
        seed,
        max_steps,
        window,
        placeholder,
        frameskip,
        max_obs,
        dt,
        damping,
        contact_force,
        contact_margin,
        **map_kwargs,
    ):
        self.map_generator = map_generator

        env = camar_v0(
            map_generator,
            window=window,
            placeholder=placeholder,
            max_steps=max_steps,
            frameskip=frameskip,
            max_obs=max_obs,
            dt=dt,
            damping=damping,
            contact_force=contact_force,
            contact_margin=contact_margin,
            **map_kwargs,
        )

        return super()._build_env(env)



from camar.wrappers.base import GymnaxWrapper
from camar.wrappers.craftax import (
    BatchEnvWrapper,
    AutoResetEnvWrapper,
    OptimisticResetVecEnvWrapper,
    LogWrapper,
)

__all__ = [
    "GymnaxWrapper",
    "BatchEnvWrapper",
    "AutoResetEnvWrapper",
    "OptimisticResetVecEnvWrapper",
    "LogWrapper",
]



import jax
import jax.numpy as jnp
import pytest
from dummy_map import dummy_map

from camar import camar_v0
from camar.environment import Camar
from camar.maps import string_grid


@pytest.fixture
def key():
    return jax.random.key(0)


@pytest.fixture
def landmark_pos():
    return jnp.array([[0.0, 0.0]])


@pytest.fixture
def agent_pos():
    return jnp.array([[1.0, 1.0]])


@pytest.fixture
def goal_pos():
    return jnp.array([[2.0, 2.0]])


@pytest.fixture
def dummy_map_generator(landmark_pos, agent_pos, goal_pos):
    return dummy_map(landmark_pos=landmark_pos, agent_pos=agent_pos, goal_pos=goal_pos)


def test_camar_v0_with_map_generator(key, dummy_map_generator):
    env = camar_v0(map_generator=dummy_map_generator)
    obs, state = env.reset(key)

    assert type(env) is Camar
    assert type(env.map_generator) is dummy_map
    assert env.height == 10.0
    assert env.width == 10.0
    assert env.num_landmarks == 1
    assert env.num_agents == 1
    assert jnp.allclose(state.landmark_pos, jnp.array([[0.0, 0.0]]))
    assert jnp.allclose(state.agent_pos, jnp.array([[1.0, 1.0]]))
    assert jnp.allclose(state.goal_pos, jnp.array([[2.0, 2.0]]))


def test_camar_v0_with_string_grid_string():
    env = camar_v0(
        map_generator="string_grid",
        map_str="...",
        num_agents=1,
    )

    assert type(env) is Camar
    assert type(env.map_generator) is string_grid
    assert jnp.isclose(env.height, 0.3)
    assert jnp.isclose(env.width, 0.5)
    assert env.num_landmarks == 12  # only borders
    assert env.num_agents == 1


def test_camar_v0_with_string_grid_class():
    map_generator = string_grid(map_str="...", num_agents=1)
    env = camar_v0(map_generator=map_generator)

    assert type(env) is Camar
    assert type(env.map_generator) is string_grid
    assert jnp.isclose(env.height, 0.3)
    assert jnp.isclose(env.width, 0.5)
    assert env.num_landmarks == 12  # only borders
    assert env.num_agents == 1



import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from .base_map import base_map
from .utils import get_border_landmarks, idx2pos, perlin_noise_vectorized


class caves_cont(base_map):
    def __init__(
        self,
        num_rows: int = 128,
        num_cols: int = 128,
        scale: int = 14,
        landmark_low_ratio: float = 0.55,
        landmark_high_ratio: float = 0.72,
        free_ratio: float = 0.20,
        add_borders: bool = True,
        num_agents: int = 16,
        obstacle_size: float = 0.1,
        agent_size: float = 0.2,
    ):
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.obstacle_size = obstacle_size
        self.agent_size = agent_size

        self.grid_num_rows = int(num_rows / scale)
        self.grid_num_cols = int(num_cols / scale)

        if landmark_low_ratio >= landmark_high_ratio:
            raise ValueError("0th element of landmark_ranks must be less than 1th.")

        num_cells = num_rows * num_cols
        self.landmark_ranks = (
            int(num_cells * landmark_low_ratio),
            int(num_cells * landmark_high_ratio),
        )
        self.num_landmarks = self.landmark_ranks[1] - self.landmark_ranks[0]

        self.free_rank = int(num_cells * free_ratio)

        self.num_agents = num_agents

        if add_borders:
            grain_factor = 2
            self.border_landmarks = get_border_landmarks(
                num_rows,
                num_cols,
                half_width=self.width / 2,
                half_height=self.height / 2,
                grain_factor=grain_factor,
            )
            self.num_landmarks += (num_rows + num_cols) * 2 * (grain_factor - 1)
        else:
            self.border_landmarks = jnp.empty(shape=(0, 2))

    @property
    def landmark_rad(self) -> float:
        return self.obstacle_size / 2

    @property
    def agent_rad(self):
        return self.agent_size / 2

    @property
    def goal_rad(self):
        return self.agent_rad / 2.5

    @property
    def height(self):
        return self.num_cols * self.obstacle_size

    @property
    def width(self):
        return self.num_rows * self.obstacle_size

    def reset(self, key: ArrayLike) -> tuple[Array, Array, Array, Array]:
        key_o, key_a, key_g = jax.random.split(key, 3)

        # generate perlin noise
        noise = perlin_noise_vectorized(
            key_o, self.num_cols, self.num_rows, self.grid_num_cols, self.grid_num_rows
        )

        noise = jnp.abs(noise).flatten()

        # extract landmarks
        landmark_idx_high = jnp.argpartition(noise, self.landmark_ranks[1])[
            : self.landmark_ranks[1]
        ]
        landmark_idx_low = jnp.argpartition(noise, self.landmark_ranks[0])[
            : self.landmark_ranks[0]
        ]
        landmark_idx = jnp.setdiff1d(
            landmark_idx_high,
            landmark_idx_low,
            size=self.landmark_ranks[1] - self.landmark_ranks[0],
        )
        landmark_idx_x, landmark_idx_y = jnp.divmod(landmark_idx, self.num_cols)

        landmark_pos = idx2pos(
            landmark_idx_x, landmark_idx_y, self.obstacle_size, self.height, self.width
        )

        # add borders
        landmark_pos = jnp.vstack((landmark_pos, self.border_landmarks))

        # extract free pos
        free_idx = jnp.argpartition(noise, self.free_rank)[: self.free_rank]
        free_idx_x, free_idx_y = jnp.divmod(free_idx, self.num_cols)

        free_pos = idx2pos(
            free_idx_x, free_idx_y, self.obstacle_size, self.height, self.width
        )

        # generate agents
        agent_pos = jax.random.choice(
            key_a, free_pos, shape=(self.num_agents,), replace=False
        )

        # generate goals
        goal_pos = jax.random.choice(
            key_g, free_pos, shape=(self.num_agents,), replace=False
        )

        return (
            key_g,
            landmark_pos,
            agent_pos,
            goal_pos,
        )  # return key_g because of lifelong



import jax
import jax.numpy as jnp

from camar.maps import batched_string_grid

MAP_STR_1 = """
.....#.....
.....#.....
...........
...........
.....#.....
.....#.....
"""

MAP_STR_2 = """
...........
..#######..
..#.....#..
..#.....#..
..#######..
...........
"""

MAP_STR_BATCH = [MAP_STR_1, MAP_STR_2]

MAP_STR_SQUARE_1 = """
.....
.###.
.#.#.
.###.
.....
"""  # 5x5

MAP_STR_SQUARE_2 = """
.#.#.
#.#.#
.#.#.
#.#.#
.#.#.
"""  # 5x5

MAP_STR_SQUARE_BATCH = [MAP_STR_SQUARE_1, MAP_STR_SQUARE_2]


class TestBatchedStringGrid:
    def test_map_creation(self):
        map_gen = batched_string_grid(map_str_batch=MAP_STR_BATCH, num_agents=2)
        assert map_gen is not None
        assert map_gen.num_agents == 2
        assert map_gen.batch_size == len(MAP_STR_SQUARE_BATCH)
        assert map_gen.num_landmarks > 0

    def test_map_reset(self):
        num_agents = 2
        map_gen = batched_string_grid(
            map_str_batch=MAP_STR_BATCH, num_agents=num_agents
        )
        key = jax.random.key(0)

        keys_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape == (map_gen.num_landmarks, 2)
        assert agent_pos.shape == (num_agents, 2)
        assert goal_pos.shape == (num_agents, 2)

    def test_map_creation_with_specific_agent_goal_pos(self):
        agent_idx_batch = [
            jnp.array([[2, 2]]),  # For MAP_STR_SQUARE_1
            jnp.array([[0, 0]]),  # For MAP_STR_SQUARE_2
        ]
        goal_idx_batch = [
            jnp.array([[0, 0]]),
            jnp.array([[0, 2]]),
        ]

        map_gen = batched_string_grid(
            map_str_batch=MAP_STR_SQUARE_BATCH,
            num_agents=1,
            agent_idx_batch=agent_idx_batch,
            goal_idx_batch=goal_idx_batch,
            add_border=False,
            remove_border=False,
        )
        key = jax.random.key(1)
        _, _, agent_pos, goal_pos = map_gen.reset(key)

        assert agent_pos is not None
        assert goal_pos is not None
        assert agent_pos.shape == (1, 2)
        assert goal_pos.shape == (1, 2)



import jax
import jax.numpy as jnp

from .const import PREGEN_DEVICE


def map_str2array(
    map_str, remove_border, add_border, preprocess=lambda map_array: map_array
):
    map_array = []
    for line in map_str.split("\n"):
        line_array = []
        if set(line) == set(" ") or not line:
            continue
        for char in line.strip():
            if char in set("@*#"):
                line_array.append(1)
            else:
                line_array.append(0)
        map_array.append(line_array)

    map_array = jnp.array(map_array, device=PREGEN_DEVICE)

    map_array = preprocess(map_array)

    if remove_border:
        map_array = map_array[1:-1, 1:-1]

    if add_border:
        map_array = jnp.pad(
            map_array, pad_width=[(1, 1), (1, 1)], constant_values=[(1, 1), (1, 1)]
        )

    return map_array


def idx2pos(idx_x, idx_y, obstacle_size, height, width):
    coord_x = idx_x * obstacle_size - height / 2 + obstacle_size / 2
    coord_y = idx_y * obstacle_size - width / 2 + obstacle_size / 2

    return jnp.stack((coord_y, coord_x), axis=1)


def parse_map_array(map_array, obstacle_size, free_pos_array=None):
    num_rows, num_cols = map_array.shape

    map_idx_rows, map_idx_cols = jnp.meshgrid(
        jnp.arange(num_cols, device=PREGEN_DEVICE),
        jnp.arange(num_rows, device=PREGEN_DEVICE),
    )

    height = num_rows * obstacle_size
    width = num_cols * obstacle_size

    # obstacles
    landmark_idx_x, landmark_idx_y = jnp.nonzero(map_array)
    landmark_pos = idx2pos(landmark_idx_x, landmark_idx_y, obstacle_size, height, width)

    # free cells
    map_idx = jnp.stack((map_idx_cols, map_idx_rows), axis=2).reshape(
        -1, 2
    )  # for random agent and goal positions
    if free_pos_array is None:
        is_free = ~map_array.flatten().astype(jnp.bool_)
    else:
        map_array_free = ~map_array.flatten().astype(jnp.bool_)
        is_free = ~free_pos_array.flatten().astype(jnp.bool_)
        is_free = jnp.logical_and(map_array_free, is_free)

    free_idx = map_idx[is_free, :]
    free_pos = idx2pos(free_idx[:, 0], free_idx[:, 1], obstacle_size, height, width)

    return landmark_pos, free_pos, height, width


# for batched string grid
def pad_placeholder(pos, num_pos, placeholder=-100.0):
    return jnp.concatenate(
        (
            pos,
            jnp.full(
                shape=(num_pos - pos.shape[0], pos.shape[1]),
                fill_value=placeholder,
                dtype=pos.dtype,
                device=PREGEN_DEVICE,
            ),
        ),
        axis=0,
    )


def random_truncate(pos, num_pos, key=jax.random.key(0)):
    return jax.random.choice(key, pos, shape=(num_pos,), replace=False)


def check_pos(map_array, pos):
    agent_cells = map_array[pos[:, 0], pos[:, 1]]
    return ~agent_cells.any()


# random_grid
def get_border_landmarks(num_rows, num_cols, half_width, half_height, grain_factor):
    top_wall = jnp.stack(
        (
            jnp.linspace(
                -half_width,  # start
                half_width,  # end
                num_rows * (grain_factor - 1),  # num points
                endpoint=False,
            ),
            jnp.full(
                (num_rows * (grain_factor - 1),),  # num points
                half_height,
            ),  # y coord of the top wall
        ),
        axis=-1,
    )
    right_wall = jnp.stack(
        (
            jnp.full(
                (num_cols * (grain_factor - 1),),  # num points
                half_width,
            ),  # x coord of the right wall
            jnp.linspace(
                half_height,  # start
                -half_height,  # end
                num_cols * (grain_factor - 1),  # num points
                endpoint=False,
            ),
        ),
        axis=-1,
    )
    bottom_wall = jnp.stack(
        (
            jnp.linspace(
                half_width,  # start
                -half_width,  # end
                num_rows * (grain_factor - 1),  # num points
                endpoint=False,
            ),
            jnp.full(
                (num_rows * (grain_factor - 1),),  # num points
                -half_height,
            ),  # y coord of the bottom wall
        ),
        axis=-1,
    )
    left_wall = jnp.stack(
        (
            jnp.full(
                (num_cols * (grain_factor - 1),),  # num points
                -half_width,
            ),  # x coord of the left wall
            jnp.linspace(
                -half_height,  # start
                half_height,  # end
                num_cols * (grain_factor - 1),  # num points
                endpoint=False,
            ),
        ),
        axis=-1,
    )
    return jnp.concatenate([top_wall, right_wall, left_wall, bottom_wall])


# movingai
def delete_movingai_header(map_str):
    map_str = map_str.replace("\r", "").split("\n")[4:]
    map_str = "\n".join(map_str)
    return map_str


def get_movingai(map_names):
    import io
    import os
    import zipfile

    import requests

    movingai_path = ".cache/movingai/"
    if not os.path.exists(movingai_path):
        os.makedirs(movingai_path)

    zip_files = {}
    map_str_batch = []
    for collection, map_name in map(lambda x: x.split("/"), map_names):
        path_to_map = f".cache/movingai/{collection}/{map_name}.map"
        if not os.path.exists(path_to_map):
            if collection not in zip_files:
                url_collection = (
                    f"https://movingai.com/benchmarks/{collection}/{collection}-map.zip"
                )
                response = requests.get(url_collection)
                zip_file = io.BytesIO(response.content)

                z = zipfile.ZipFile(zip_file, "r")
                zip_files[collection] = z

            z = zip_files[collection]
            map_file_name = f"{map_name}.map"
            if map_file_name not in z.namelist():
                raise ValueError(f"there is no {map_file_name} in {url_collection}.")

            with z.open(map_file_name, "r") as f:
                map_str = f.read().decode()
                map_str_batch.append(delete_movingai_header(map_str))

                os.makedirs(os.path.dirname(path_to_map), exist_ok=True)
                with open(path_to_map, "w") as output_f:
                    output_f.write(map_str)
        else:
            with open(path_to_map, "r") as f:
                map_str = f.read()
                map_str_batch.append(delete_movingai_header(map_str))

    for collection, z in zip_files.items():
        z.close()

    return map_str_batch


def detect_edges(img, low_thr, mode="same"):
    sobel_x = jnp.array([[-1, 0, 1], [-2, 0, 2], [-1, 0, 1]], device=PREGEN_DEVICE)

    sobel_y = jnp.array([[-1, -2, -1], [0, 0, 0], [1, 2, 1]], device=PREGEN_DEVICE)

    edges_x = jax.scipy.signal.convolve2d(img, sobel_x, mode=mode)
    edges_y = jax.scipy.signal.convolve2d(img, sobel_y, mode=mode)

    edges = jnp.sqrt(edges_x**2 + edges_y**2)

    return edges > low_thr


def fade(t):
    return 6 * t**5 - 15 * t**4 + 10 * t**3


def lerp(a, b, t):
    return a + t * (b - a)


def generate_gradients(key, grid_width, grid_height):
    angles = jax.random.uniform(
        key, shape=(grid_width + 1, grid_height + 1), minval=0.0, maxval=2.0 * jnp.pi
    )
    gradients = jnp.dstack((jnp.cos(angles), jnp.sin(angles)))
    return gradients


def perlin_noise_vectorized(key, width, height, grid_width, grid_height):
    gradients = generate_gradients(key, grid_width, grid_height)

    xs = jnp.linspace(0, grid_width, width, endpoint=False)
    ys = jnp.linspace(0, grid_height, height, endpoint=False)
    x, y = jnp.meshgrid(xs, ys)

    x0 = x.astype(int)
    y0 = y.astype(int)
    x1 = x0 + 1
    y1 = y0 + 1

    sx = fade(x - x0)
    sy = fade(y - y0)

    # Top-left
    n00 = (x - x0) * gradients[x0, y0, 0] + (y - y0) * gradients[x0, y0, 1]
    # Top-right
    n10 = (x - x1) * gradients[x1, y0, 0] + (y - y0) * gradients[x1, y0, 1]
    # Bottom-left
    n01 = (x - x0) * gradients[x0, y1, 0] + (y - y1) * gradients[x0, y1, 1]
    # Bottom-right
    n11 = (x - x1) * gradients[x1, y1, 0] + (y - y1) * gradients[x1, y1, 1]

    # Interpolate the dot products along the x-direction.
    nx0 = lerp(n00, n10, sx)
    nx1 = lerp(n01, n11, sx)

    # Interpolate the results along the y-direction to get the final noise.
    nxy = lerp(nx0, nx1, sy)

    return nxy



import jax
import jax.numpy as jnp

from camar.utils import Box


def test_box_sample():
    key = jax.random.key(0)
    low = -5.0
    high = 5.0
    shape = (4,)
    dtype = jnp.float32
    box_space = Box(low=low, high=high, shape=shape, dtype=dtype)

    sample = box_space.sample(key)

    assert sample.shape == shape
    assert jnp.isdtype(sample.dtype, dtype)
    assert jnp.all(sample >= low)
    assert jnp.all(sample < high)



from typing import Optional, Callable

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from .base_map import base_map
from .const import ENV_DEVICE
from .utils import idx2pos, map_str2array, parse_map_array, random_truncate


class string_grid(base_map):
    def __init__(
        self,
        map_str: str,
        free_pos_str: Optional[str] = None,
        agent_idx: Optional[ArrayLike] = None,
        goal_idx: Optional[ArrayLike] = None,
        num_agents: int = 10,
        random_agents: bool = True,
        random_goals: bool = True,
        remove_border: bool = False,
        add_border: bool = True,
        obstacle_size: float = 0.1,
        agent_size: float = 0.04,
        max_free_pos: Optional[int] = None,
        map_array_preprocess: Callable[[ArrayLike], Array] = lambda map_array: map_array,
        free_pos_array_preprocess: Callable[[ArrayLike], Array] = lambda free_pos_array: free_pos_array,
    ) -> base_map:
        if agent_idx is not None:
            num_agents = agent_idx.shape[0]
        if goal_idx is not None:
            num_agents = goal_idx.shape[0]

        self.num_agents = num_agents
        self.obstacle_size = obstacle_size
        self.agent_size = agent_size

        map_array = map_str2array(
            map_str, remove_border, add_border, map_array_preprocess
        )

        if free_pos_str is not None:
            free_pos_array = map_str2array(
                free_pos_str, remove_border, add_border, free_pos_array_preprocess
            )
        else:
            free_pos_array = None

        if agent_idx is not None:
            if remove_border:
                agent_idx -= 1

            if add_border:
                agent_idx += 1

            agent_cells = map_array[agent_idx[:, 0], agent_idx[:, 1]]
            assert ~agent_cells.any(), f"agent_idx must be free. got {agent_cells}"

        if goal_idx is not None:
            if remove_border:
                goal_idx -= 1

            if add_border:
                goal_idx += 1

            goal_cells = map_array[goal_idx[:, 0], goal_idx[:, 1]]
            assert ~goal_cells.any(), f"goal_idx must be free. got {goal_cells}"

        self.landmark_pos, free_pos, self._height, self._width = parse_map_array(
            map_array, obstacle_size, free_pos_array
        )
        self.landmark_pos = self.landmark_pos.to_device(ENV_DEVICE)

        if max_free_pos is not None:
            free_pos = random_truncate(free_pos, max_free_pos)
        free_pos = free_pos.to_device(ENV_DEVICE)

        if agent_idx is not None:
            agent_pos = idx2pos(
                agent_idx[:, 0], agent_idx[:, 1], obstacle_size, self.height, self.width
            )
            self.generate_agents = lambda key: agent_pos
        elif random_agents:
            self.generate_agents = lambda key: jax.random.choice(
                key, free_pos, shape=(self.num_agents,), replace=False
            )
        else:
            agent_pos = jax.random.choice(
                jax.random.key(0), free_pos, shape=(self.num_agents,), replace=False
            )
            self.generate_agents = lambda key: agent_pos

        if goal_idx is not None:
            goal_pos = idx2pos(
                goal_idx[:, 0], goal_idx[:, 1], obstacle_size, self.height, self.width
            )
            self.generate_goals = lambda key: goal_pos
        elif random_goals:
            self.generate_goals = lambda key: jax.random.choice(
                key, free_pos, shape=(self.num_agents,), replace=False
            )
            self.generate_goals_lifelong = jax.vmap(
                lambda key: jax.random.choice(key, free_pos), in_axes=[0]
            )  # 1 key = 1 goal
        else:
            goal_pos = jax.random.choice(
                jax.random.key(1), free_pos, shape=(self.num_agents,), replace=False
            )
            self.generate_goals = lambda key: goal_pos

        self.num_landmarks = self.landmark_pos.shape[0]

    @property
    def landmark_rad(self) -> float:
        return self.obstacle_size / 2

    @property
    def agent_rad(self):
        return self.agent_size / 2

    @property
    def goal_rad(self):
        return self.agent_rad / 2.5

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    def reset(self, key: ArrayLike) -> tuple[Array, Array, Array, Array]:
        key_a, key_g = jax.random.split(key, 2)

        # generate agents
        agent_pos = self.generate_agents(key_a)

        # generate goals
        goal_pos = self.generate_goals(key_g)

        return (
            key_g,
            self.landmark_pos,
            agent_pos,
            goal_pos,
        )  # return key_g because of lifelong

    def reset_lifelong(self, key) -> tuple[Array, Array, Array, Array]:
        key_a, key_g = jax.random.split(key, 2)

        # generate agents
        agent_pos = self.generate_agents(key_a)

        # generate goals
        # key for each goal
        key_g = jax.random.split(key_g, self.num_agents)

        goal_pos = self.generate_goals_lifelong(key_g)

        return key_g, self.landmark_pos, agent_pos, goal_pos

    def update_goals(
        self, keys: ArrayLike, goal_pos: ArrayLike, to_update: ArrayLike
    ) -> tuple[Array, Array]:
        new_keys = jax.vmap(jax.random.split, in_axes=[0, None])(keys, 1)[:, 0]
        new_keys = jnp.where(to_update, new_keys, keys)

        new_goal_pos = self.generate_goals_lifelong(new_keys)

        return new_keys, new_goal_pos



# From Craftax Baselines https://github.com/MichaelTMatthews/Craftax_Baselines.git


class GymnaxWrapper(object):
    """Base class for Gymnax wrappers."""

    def __init__(self, env):
        self._env = env

    # provide proxy access to regular attributes of wrapped object
    def __getattr__(self, name):
        return getattr(self._env, name)



from functools import partial
from typing import Optional

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from camar.maps import random_grid
from camar.maps.base_map import base_map
from camar.utils import Box, State


class Camar:
    def __init__(
        self,
        map_generator: base_map = random_grid(),
        lifelong: bool = False,
        window: Optional[float] = None,
        placeholder: float = 0.0,
        max_steps: int = 100,
        frameskip: int = 1,
        max_obs: Optional[int] = None,
        pos_shaping_factor: Optional[float] = None,
        dt: float = 0.01,
        damping: float = 0.25,
        contact_force: float = 500,
        contact_margin: float = 0.001,
        **kwargs,
    ):
        self.device = str(jax.devices()[0])

        self.map_generator = map_generator

        self.frameskip = frameskip

        self.placeholder = placeholder

        self.height = map_generator.height
        self.width = map_generator.width
        self.landmark_rad = map_generator.landmark_rad
        self.agent_rad = map_generator.agent_rad
        self.goal_rad = map_generator.goal_rad

        if pos_shaping_factor is None:
            self.pos_shaping_factor = 0.04 / self.agent_rad
        else:
            self.pos_shaping_factor = pos_shaping_factor

        if window is None:
            self.window = max(self.landmark_rad, self.agent_rad) * 2.5
        else:
            self.window = window

        if max_obs is not None:
            self.max_obs = max_obs
        else:
            self.max_obs = int(self.window / self.landmark_rad)**2 # for partial observability

        self.max_obs = min(self.max_obs, self.num_entities - 1)

        self.action_size = 2
        self.observation_size = self.max_obs * 2 + 2

        self.action_spaces = Box(low=-1.0, high=1.0, shape=(self.num_agents, self.action_size))
        self.observation_spaces = Box(-jnp.inf, jnp.inf, shape=(self.num_agents, self.observation_size))

        # Environment parameters
        self.max_steps = max_steps
        self.dt = dt
        self.step_dt = self.dt * (self.frameskip + 1) # for metrics
        self.max_time = self.max_steps * self.step_dt # for metrics

        self.mass = kwargs.get("mass", 1.0)
        self.accel = kwargs.get("accel", 5.0)
        self.max_speed = kwargs.get("max_speed", -1)
        self.u_noise = kwargs.get("u_noise", 0)

        self.damping = damping
        self.contact_force = contact_force
        self.contact_margin = contact_margin

        # lifelong
        self.map_reset = map_generator.reset_lifelong if lifelong else map_generator.reset
        self.update_goals = map_generator.update_goals if lifelong else lambda keys, goal_pos, to_update: (keys, goal_pos)

    @property
    def num_agents(self) -> int:
        return self.map_generator.num_agents

    @property
    def num_landmarks(self) -> int:
        return self.map_generator.num_landmarks

    @property
    def num_entities(self) -> int:
        return self.num_agents + self.num_landmarks

    def step(self, key: ArrayLike, state: State, actions: ArrayLike) -> tuple[Array, State, Array, Array, dict]:
        # actions = (num_agents, 2)
        u = self.accel * actions

        key, key_w = jax.random.split(key)

        old_goal_dist = jnp.linalg.norm(state.agent_pos - state.goal_pos, axis=-1) # (num_agents, )

        def frameskip(scan_state, _):
            key, state, u, is_collision = scan_state

            key, _key = jax.random.split(key)
            state, is_collision = self._world_step(_key, state, u, is_collision)

            return (key, state, u, is_collision), None

        is_collision = jnp.zeros(shape=(self.num_agents, ), dtype=jnp.int32)

        (key, state, u, is_collision), _ = jax.lax.scan(frameskip, init=(key_w, state, u, is_collision), xs=None, length=self.frameskip + 1)

        is_collision = is_collision >= 1

        goal_dist = jnp.linalg.norm(state.agent_pos - state.goal_pos, axis=-1) # (num_agents, )
        on_goal = goal_dist < self.goal_rad

        # done = jnp.full((self.num_agents, ), state.step >= self.max_steps)

        done = jnp.logical_or(state.step >= self.max_steps, on_goal.all(axis=-1))

        # terminated = on_goal.all(axis=-1)
        # truncated = state.step >= self.max_steps

        reward = self.get_reward(is_collision, goal_dist, old_goal_dist)

        goal_keys, goal_pos = self.update_goals(state.goal_keys, state.goal_pos, on_goal)

        just_arrived = jnp.logical_not(state.on_goal) & on_goal
        current_time = (state.step + 1) * self.step_dt
        time_to_reach_goal = jnp.where(just_arrived, current_time, state.time_to_reach_goal)
        num_collisions = state.num_collisions + is_collision.astype(jnp.int32)

        state = state.replace(
            goal_pos = goal_pos,
            is_collision = is_collision,
            step = state.step + 1,
            goal_keys = goal_keys,
            on_goal = on_goal,
            time_to_reach_goal = time_to_reach_goal,
            num_collisions = num_collisions,
        )

        obs = self.get_obs(state)

        return obs, state, reward, done, {}

    def reset(self, key: ArrayLike) -> tuple[Array, State]:
        """Initialise with random positions"""

        goal_keys, landmark_pos, agent_pos, goal_pos = self.map_reset(key)

        # reward = self.get_reward(agent_pos, all_landmark_pos, goal_pos)

        goal_dist = jnp.linalg.norm(agent_pos - goal_pos, axis=-1)
        on_goal = goal_dist < self.goal_rad

        state = State(
            agent_pos = agent_pos,
            agent_vel = jnp.zeros((self.num_agents, 2)),
            goal_pos = goal_pos,
            landmark_pos = landmark_pos,
            is_collision = jnp.full((self.num_agents, ), False, dtype=jnp.bool_),
            step = 0,
            on_goal = on_goal,
            time_to_reach_goal = jnp.full((self.num_agents, ), self.max_time),
            num_collisions = jnp.zeros((self.num_agents, ), dtype=jnp.int32),
            goal_keys = goal_keys,
        )

        obs = self.get_obs(state)

        return obs, state

    @partial(jax.vmap, in_axes=[None, 0, None])
    def get_dist(self, a_pos: ArrayLike, p_pos: ArrayLike) -> Array:
        return jnp.linalg.norm(a_pos - p_pos, axis=-1)

    def get_obs(self, state: State) -> Array:
        agent_pos = state.agent_pos
        goal_pos = state.goal_pos
        landmark_pos = state.landmark_pos

        objects = jnp.vstack((agent_pos, landmark_pos)) # (num_objects, 2)

        # (1, num_objects, 2) - (num_agents, 1, 2) -> (num_agents, num_objecst, 2)
        ego_objects = objects[None, :, :] - agent_pos[:, None, :]

        # (num_agents, num_objecst, 2) -> (num_agents, num_objecst)
        dists = jnp.linalg.norm(ego_objects, axis=-1)
        nearest_dists, nearest_ids = jax.lax.top_k(- dists, self.max_obs + 1) # (num_agents, self.max_obs + 1)
        # remove zero dists (nearest is the agent itself) -> (num_agents, self.max_obs)
        nearest_ids = nearest_ids[:, 1:]
        nearest_dists = -nearest_dists[:, 1:]

        nearest_ego_objects = ego_objects[jnp.arange(self.num_agents)[:, None], nearest_ids] # (num_agents, self.max_obs, 2)
        nearest_rad = jnp.where(nearest_ids < self.num_agents,
                                self.agent_rad,
                                self.landmark_rad) # (num_agents, self.max_obs)

        obs_dists_coeff = (self.agent_rad + nearest_rad) / nearest_dists # (num_agents, self.max_obs)
        obs_dists = jnp.linalg.norm(nearest_ego_objects * (1.0 - obs_dists_coeff)[:, :, None], axis=-1)  # (num_agents, self.max_obs)

        obs_coeff = (self.window + nearest_rad) / nearest_dists # (num_agents, self.max_obs)
        obs = nearest_ego_objects * (1.0 - obs_coeff)[:, :, None] # (num_agents, self.max_obs, 2)
        obs_norm = obs / (self.window - self.agent_rad)

        obs_norm = jnp.where(obs_dists[:, :, None] < self.window,
                             obs_norm,
                             0.0)  # (num_agents, self.max_obs, 2)

        ego_goal = goal_pos - agent_pos # [num_agents, 2]

        goal_dist = jnp.linalg.norm(ego_goal, axis=-1)

        ego_goal_norm = jnp.where(goal_dist[:, None] > 1.0,
                                  ego_goal / goal_dist[:, None],
                                  ego_goal)

        obs = jnp.concatenate((ego_goal_norm[:, None, :], obs_norm), axis=1) # (num_agents, self.max_obs + goal, 2)

        return obs.reshape(self.num_agents, self.observation_size)

    def get_reward(self, is_collision: ArrayLike, goal_dist: ArrayLike, old_goal_dist: ArrayLike) -> Array:

        on_goal = goal_dist < self.goal_rad

        r = 0.5 * on_goal.astype(jnp.float32) - 1.0 * is_collision.astype(jnp.float32) + self.pos_shaping_factor * (old_goal_dist - goal_dist)
        return r.reshape(-1, 1)

    def _world_step(self, key: ArrayLike, state: State, u: ArrayLike, is_collision: ArrayLike) -> tuple[State, Array]:

        agent_force = self._add_noise(key, u)

        # apply collision forces
        agent_force, is_collision = self._apply_environment_force(agent_force, is_collision, state)

        # integrate state
        state = self._integrate_state(agent_force, state)

        return state, is_collision

    def _add_noise(self, key: ArrayLike, u: ArrayLike) -> Array:
        noise = jax.random.normal(key, shape=u.shape) * self.u_noise
        return u + noise

    def _integrate_state(self, force: ArrayLike, state: State) -> State:
        """integrate physical state"""
        pos = state.agent_pos
        vel = state.agent_vel

        pos += vel * self.dt
        vel = vel * (1 - self.damping)

        vel += (force / self.mass) * self.dt

        speed = jnp.linalg.norm(vel, axis=-1, keepdims=True)
        over_max = vel / speed * self.max_speed

        vel = jnp.where((speed > self.max_speed) & (self.max_speed >= 0), over_max, vel)

        state = state.replace(
            agent_pos = pos,
            agent_vel = vel,
        )

        return state

    def _apply_environment_force(self, agent_force: ArrayLike, is_collision: ArrayLike, state: State) -> tuple[Array, Array]:

        # agent - agent collisions
        agent_idx_i, agent_idx_j = jnp.triu_indices(self.num_agents, k=1)
        agent_forces, is_collision_agents = self._get_collision_force(state.agent_pos[agent_idx_i], state.agent_pos[agent_idx_j], self.agent_rad + self.agent_rad) # (num_agents * (num_agents - 1) / 2, 2)

        is_collision = is_collision.at[agent_idx_i].add(is_collision_agents)
        is_collision = is_collision.at[agent_idx_j].add(is_collision_agents)

        agent_force = agent_force.at[agent_idx_i].add(agent_forces)
        agent_force = agent_force.at[agent_idx_j].add(- agent_forces)

        # agent - landmark collisions
        agent_idx = jnp.repeat(jnp.arange(self.num_agents), self.num_landmarks)
        landmark_idx = jnp.tile(jnp.arange(self.num_landmarks), self.num_agents)
        landmark_forces, is_collision_landmarks = self._get_collision_force(state.agent_pos[agent_idx], state.landmark_pos[landmark_idx], self.agent_rad + self.landmark_rad) # (num_agents * num_landmarks, 2)

        is_collision = is_collision.at[agent_idx].add(is_collision_landmarks)

        agent_force = agent_force.at[agent_idx].add(landmark_forces)

        return agent_force, is_collision

    @partial(jax.vmap, in_axes=[None, 0, 0, None])
    def _get_collision_force(self, pos_a: ArrayLike, pos_b: ArrayLike, min_dist: float) -> tuple[Array, Array]:
        delta_pos = pos_a - pos_b

        dist = jnp.linalg.norm(delta_pos, axis=-1)

        # softmax penetration
        k = self.contact_margin
        penetration = jnp.logaddexp(0, - (dist - min_dist) / k) * k
        force = self.contact_force * delta_pos / jax.lax.select(dist > 0, dist, jnp.full(dist.shape, 1e-8)) * penetration
        is_collision = (dist < min_dist).astype(jnp.int32)
        return force, is_collision



import colorsys


def hex_to_hsl(hex_color):
    hex_color = hex_color.lstrip('#')
    r = int(hex_color[0:2], 16)
    g = int(hex_color[2:4], 16)
    b = int(hex_color[4:6], 16)

    r /= 255.0
    g /= 255.0
    b /= 255.0

    hue, lightness, saturation = colorsys.rgb_to_hls(r, g, b)

    hue = round(hue * 360)
    saturation = round(saturation * 100)
    lightness = round(lightness * 100)

    return hue, saturation, lightness



from abc import ABC, abstractmethod
from typing import Tuple

from jax import Array
from jax.typing import ArrayLike


class base_map(ABC):
    @property
    @abstractmethod
    def landmark_rad(self) -> float:
        pass

    @property
    @abstractmethod
    def agent_rad(self) -> float:
        pass

    @property
    @abstractmethod
    def goal_rad(self) -> float:
        pass

    @property
    @abstractmethod
    def height(self) -> float:
        pass

    @property
    @abstractmethod
    def width(self) -> float:
        pass

    def reset(
        self, key: ArrayLike
    ) -> Tuple[
        Array, Array, Array, Array
    ]:  # Tuple[jax.random.key, landmark_pos, agent_pos, goal_pos]
        raise NotImplementedError(
            f"{self.__class__.__name__}.reset is not implemented. Must be implemented if lifelong=False."
        )

    def reset_lifelong(
        self, key: ArrayLike
    ) -> Tuple[
        Array, Array, Array, Array
    ]:  # Tuple[jax.random.key, landmark_pos, agent_pos, goal_pos]
        raise NotImplementedError(
            f"{self.__class__.__name__}.reset_lifelong is not implemented. Must be implemented if lifelong=True."
        )

    def update_goals(
        self, keys: ArrayLike, goal_pos: ArrayLike, to_update: ArrayLike
    ) -> Tuple[Array, Array]:  # Tuple[jax.random.key, goal_pos]
        raise NotImplementedError(
            f"{self.__class__.__name__}.update_goals is not implemented. Must be implemented if lifelong=True."
        )



from functools import partial
from typing import Tuple

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from .base_map import base_map
from .utils import get_border_landmarks


class random_grid(base_map):
    def __init__(
        self,
        num_rows: int = 20,
        num_cols: int = 20,
        obstacle_density: float = 0.2,
        num_agents: int = 32,
        grain_factor: int = 3,
        obstacle_size: float = 0.4,
    ) -> base_map:
        self.num_rows = num_rows
        self.num_cols = num_cols
        self.obstacle_density = obstacle_density
        self.num_agents = num_agents
        self.grain_factor = grain_factor
        self.obstacle_size = obstacle_size

        # helpful params
        self.num_obstacles = int(obstacle_density * num_rows * num_cols)
        self.num_landmarks = self.num_obstacles * 4 * (self.grain_factor - 1)
        self.num_landmarks += (
            (num_rows + num_cols) * 2 * (self.grain_factor - 1)
        )  # adding borders

        half_width = self.width / 2
        half_height = self.height / 2

        x_coords = jnp.linspace(
            -half_width + self.obstacle_size / 2,  # start
            half_width - self.obstacle_size / 2,  # end
            num_rows,  # map num_rows
        )
        y_coords = jnp.linspace(
            -half_height + self.obstacle_size / 2,  # start
            half_height - self.obstacle_size / 2,  # end
            num_cols,  # map num_cols
        )

        # coordinates for sampling
        self.map_coordinates = jnp.stack(
            jnp.meshgrid(x_coords, y_coords), axis=-1
        ).reshape(-1, 2)  # cell centers of the whole map
        self.border_landmarks = get_border_landmarks(
            num_rows, num_cols, half_width, half_height, self.grain_factor
        )

    @property
    def landmark_rad(self) -> float:
        return self.obstacle_size / (2 * (self.grain_factor - 1))

    @property
    def agent_rad(self):
        return (self.obstacle_size - 2 * self.landmark_rad) * 0.25

    @property
    def goal_rad(self):
        return self.agent_rad / 2.5

    @property
    def height(self):
        return self.num_cols * self.obstacle_size

    @property
    def width(self):
        return self.num_rows * self.obstacle_size

    def reset(self, key: ArrayLike) -> Tuple[Array, Array, Array, Array]:
        permuted_pos = jax.random.permutation(key, self.map_coordinates)

        agent_pos = jax.lax.dynamic_slice(
            permuted_pos,  # [0 : num_agents, 0 : 2]
            start_indices=(0, 0),
            slice_sizes=(self.num_agents, 2),
        )

        obstacle_pos = jax.lax.dynamic_slice(
            permuted_pos,  # [num_agents : num_agents + num_obstacles, 0 : 2]
            start_indices=(self.num_agents, 0),
            slice_sizes=(self.num_obstacles, 2),
        )

        goal_pos = jax.lax.dynamic_slice(
            permuted_pos,  # [num_agents + num_obstacles : 2 * num_agents + num_obstacles, 0 : 2]
            start_indices=(self.num_agents + self.num_obstacles, 0),
            slice_sizes=(self.num_agents, 2),
        )

        landmark_pos = self.get_landmarks(
            obstacle_pos, self.grain_factor, self.obstacle_size
        ).reshape(-1, 2)

        all_landmark_pos = jnp.concatenate(
            [
                landmark_pos,
                self.border_landmarks,
            ],
        )

        return key, all_landmark_pos, agent_pos, goal_pos

    @partial(jax.vmap, in_axes=[None, 0, None, None], out_axes=1)
    def get_landmarks(self, obstacle, grain_factor, obstacle_size):
        left_x, down_y = obstacle - obstacle_size / 2
        right_x, up_y = obstacle + obstacle_size / 2

        up_landmarks = jnp.stack(
            (
                jnp.linspace(left_x, right_x, grain_factor - 1, endpoint=False),
                jnp.full((grain_factor - 1,), up_y),
            ),
            axis=-1,
        )
        right_landmarks = jnp.stack(
            (
                jnp.full((grain_factor - 1,), right_x),
                jnp.linspace(up_y, down_y, grain_factor - 1, endpoint=False),
            ),
            axis=-1,
        )
        down_landmarks = jnp.stack(
            (
                jnp.linspace(right_x, left_x, grain_factor - 1, endpoint=False),
                jnp.full((grain_factor - 1,), down_y),
            ),
            axis=-1,
        )
        left_landmarks = jnp.stack(
            (
                jnp.full((grain_factor - 1,), left_x),
                jnp.linspace(down_y, up_y, grain_factor - 1, endpoint=False),
            ),
            axis=-1,
        )

        return jnp.concatenate(
            [up_landmarks, right_landmarks, down_landmarks, left_landmarks]
        )



# From Craftax Baselines https://github.com/MichaelTMatthews/Craftax_Baselines.git with few modifications: deleted env_params logic

from functools import partial
from typing import Any, Union

import jax
import jax.numpy as jnp
from flax import struct
from jax.typing import ArrayLike

from camar.wrappers import GymnaxWrapper


class BatchEnvWrapper(GymnaxWrapper):
    """Batches reset and step functions"""

    def __init__(self, env, num_envs: int):
        super().__init__(env)

        self.num_envs = num_envs

        self.reset_fn = jax.vmap(self._env.reset, in_axes=(0, ))
        self.step_fn = jax.vmap(self._env.step, in_axes=(0, 0, 0, ))

    @partial(jax.jit, static_argnums=(0, ))
    def reset(self, rng):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, self.num_envs)
        obs, env_state = self.reset_fn(rngs, )
        return obs, env_state

    @partial(jax.jit, static_argnums=(0, ))
    def step(self, rng, state, action):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, self.num_envs)
        obs, state, reward, done, info = self.step_fn(rngs, state, action)

        return obs, state, reward, done, info


class AutoResetEnvWrapper(GymnaxWrapper):
    """Provides standard auto-reset functionality, providing the same behaviour as Gymnax-default."""

    def __init__(self, env):
        super().__init__(env)

    @partial(jax.jit, static_argnums=(0, ))
    def reset(self, key):
        return self._env.reset(key)

    @partial(jax.jit, static_argnums=(0, ))
    def step(self, rng, state, action):

        rng, _rng = jax.random.split(rng)
        obs_st, state_st, reward, done, info = self._env.step(_rng, state, action)
        done = done.reshape(1, )

        rng, _rng = jax.random.split(rng)
        obs_re, state_re = self._env.reset(_rng)

        # Auto-reset environment based on termination
        def auto_reset(done, state_re, state_st, obs_re, obs_st):
            state = jax.tree.map(
                lambda x, y: jnp.where(jnp.expand_dims(done, range(1, x.ndim)), x, y), state_re, state_st
            )
            obs = jnp.where(done[:, None, None], obs_re, obs_st)

            return obs, state

        obs, state = auto_reset(done, state_re, state_st, obs_re, obs_st)

        return obs, state, reward, done, info


class OptimisticResetVecEnvWrapper(GymnaxWrapper):
    """
    Provides efficient 'optimistic' resets.
    The wrapper also necessarily handles the batching of environment steps and resetting.
    reset_ratio: the number of environment workers per environment reset.  Higher means more efficient but a higher
    chance of duplicate resets.
    """

    def __init__(self, env, num_envs: int, reset_ratio: int):
        super().__init__(env)

        self.num_envs = num_envs
        self.reset_ratio = reset_ratio
        assert (
            num_envs % reset_ratio == 0
        ), "Reset ratio must perfectly divide num envs."
        self.num_resets = self.num_envs // reset_ratio

        self.reset_fn = jax.vmap(self._env.reset, in_axes=(0, ))
        self.step_fn = jax.vmap(self._env.step, in_axes=(0, 0, 0, ))

    @partial(jax.jit, static_argnums=(0, ))
    def reset(self, rng):
        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, self.num_envs)
        obs, env_state = self.reset_fn(rngs, )
        return obs, env_state

    @partial(jax.jit, static_argnums=(0, ))
    def step(self, rng, state, action):

        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, self.num_envs)
        obs_st, state_st, reward, done, info = self.step_fn(rngs, state, action)

        rng, _rng = jax.random.split(rng)
        rngs = jax.random.split(_rng, self.num_resets)
        obs_re, state_re = self.reset_fn(rngs)

        rng, _rng = jax.random.split(rng)
        reset_indexes = jnp.arange(self.num_resets).repeat(self.reset_ratio)

        being_reset = jax.random.choice(
            _rng,
            jnp.arange(self.num_envs),
            shape=(self.num_resets,),
            p=done,
            replace=False,
        )
        reset_indexes = reset_indexes.at[being_reset].set(jnp.arange(self.num_resets))

        obs_re = obs_re[reset_indexes]
        state_re = jax.tree.map(lambda x: x[reset_indexes], state_re)

        # Auto-reset environment based on termination
        def auto_reset(done, state_re, state_st, obs_re, obs_st):
            state = jax.tree.map(
                lambda x, y: jnp.where(jnp.expand_dims(done, range(1, x.ndim)), x, y), state_re, state_st
            )
            obs = jnp.where(done[:, None, None], obs_re, obs_st)

            return obs, state

        obs, state = auto_reset(done, state_re, state_st, obs_re, obs_st)

        return obs, state, reward, done, info


@struct.dataclass
class LogEnvState:
    env_state: Any
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int
    timestep: int


class LogWrapper(GymnaxWrapper):
    """Log the episode returns and lengths."""

    def __init__(self, env):
        super().__init__(env)

    @partial(jax.jit, static_argnums=(0, ))
    def reset(self, key: ArrayLike):
        obs, env_state = self._env.reset(key)
        state = LogEnvState(env_state, 0.0, 0, 0.0, 0, 0)
        return obs, state

    @partial(jax.jit, static_argnums=(0, ))
    def step(
        self,
        key: ArrayLike,
        state,
        action: Union[int, float],
    ):
        obs, env_state, reward, done, info = self._env.step(key, state.env_state, action)
        new_episode_return = state.episode_returns + reward
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - done),
            episode_lengths=new_episode_length * (1 - done),
            returned_episode_returns=state.returned_episode_returns * (1 - done)
            + new_episode_return * done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - done)
            + new_episode_length * done,
            timestep=state.timestep + 1,
        )
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["timestep"] = state.timestep
        info["returned_episode"] = done
        return obs, state, reward, done, info



from typing import List, Optional, Callable

import jax
import jax.numpy as jnp
from jax import Array
from jax.typing import ArrayLike

from .base_map import base_map
from .const import ENV_DEVICE
from .utils import (
    check_pos,
    idx2pos,
    map_str2array,
    pad_placeholder,
    parse_map_array,
    random_truncate,
)


class batched_string_grid(base_map):
    def __init__(
        self,
        map_str_batch: List[str],
        free_pos_str_batch: Optional[List[str]] = None,
        agent_idx_batch: Optional[List[ArrayLike]] = None,
        goal_idx_batch: Optional[List[ArrayLike]] = None,
        num_agents: int = 10,
        random_agents: bool = True,
        random_goals: bool = True,
        remove_border: bool = False,
        add_border: bool = True,
        obstacle_size: float = 0.1,
        agent_size: float = 0.06,
        max_free_pos: Optional[int] = None,
        map_array_preprocess: Callable[[ArrayLike], Array] = lambda map_array: map_array,
        free_pos_array_preprocess: Callable[[ArrayLike], Array] = lambda free_pos_array: free_pos_array,
    ) -> base_map:
        self.batch_size = len(map_str_batch)
        if agent_idx_batch is not None:
            num_agents = agent_idx_batch[0].shape[0]

            # check agent_idx_batch
            assert all(
                map(lambda agent_idx: agent_idx.shape[0] == num_agents, agent_idx_batch)
            ), "agent_idx.shape must be the same in a batch."
            assert len(agent_idx_batch) == self.batch_size

        if goal_idx_batch is not None:
            num_agents = goal_idx_batch[0].shape[0]

            # check goal_idx_batch
            assert all(
                map(lambda goal_idx: goal_idx.shape[0] == num_agents, goal_idx_batch)
            ), "goal_idx.shape must be the same in a batch."
            assert len(goal_idx_batch) == self.batch_size

        self.num_agents = num_agents
        self.obstacle_size = obstacle_size
        self.agent_size = agent_size

        map_array_batch = list(
            map(
                lambda map_str: map_str2array(
                    map_str,
                    remove_border,
                    add_border,
                    map_array_preprocess,
                ),
                map_str_batch,
            )
        )

        # check map_array_batch
        map_array_shape = map_array_batch[0].shape
        map_array_checks = map(
            lambda map_array: map_array.shape == map_array_shape, map_array_batch
        )
        assert all(map_array_checks), (
            "all maps must have the same shape."
            "Resize or provide resize map_array_preprocess."
        )

        if free_pos_str_batch is not None:
            free_pos_array_batch = list(
                map(
                    lambda free_pos_str: map_str2array(
                        free_pos_str,
                        remove_border,
                        add_border,
                        free_pos_array_preprocess,
                    ),
                    free_pos_str_batch,
                )
            )

            # check free_pos_array_batch
            free_pos_array_checks = map(
                lambda map_array, free_pos_array: map_array.shape
                == free_pos_array.shape,
                map_array_batch,
                free_pos_array_batch,
            )
            assert all(free_pos_array_checks), (
                "all free_pos must have the same shape as maps after processing."
                "Resize or provide resize free_pos_array_preprocess."
            )
        else:
            free_pos_array_batch = [None for _ in map_str_batch]

        if agent_idx_batch is not None:
            if remove_border:
                agent_idx_batch = [idx - 1 for idx in agent_idx_batch]

            if add_border:
                agent_idx_batch = [idx + 1 for idx in agent_idx_batch]

            # check if agents on free cells
            agent_checks = map(
                lambda map_array, agent_idx: check_pos(map_array, agent_idx),
                map_array_batch,
                agent_idx_batch,
            )
            assert any(agent_checks), "agent_idx must be free for each map instance."

        if goal_idx_batch is not None:
            if remove_border:
                goal_idx_batch = [idx - 1 for idx in goal_idx_batch]

            if add_border:
                goal_idx_batch = [idx + 1 for idx in goal_idx_batch]

            # check if goals on free cells
            goal_checks = map(
                lambda map_array, goal_idx: check_pos(map_array, goal_idx),
                map_array_batch,
                goal_idx_batch,
            )
            assert any(goal_checks), "goal_idx must be free for each map instance."

        self.landmark_pos_batch, free_pos_batch, height_batch, width_batch = zip(
            *map(
                lambda map_array, free_pos_array: parse_map_array(
                    map_array, obstacle_size, free_pos_array
                ),
                map_array_batch,
                free_pos_array_batch,
            )
        )

        # check height in a batch
        self._height = height_batch[0]
        assert all(map(lambda height: height == self._height, height_batch)), (
            "map height must be the same in a batch."
        )

        # check width in a batch
        self._width = width_batch[0]
        assert all(map(lambda width: width == self._width, width_batch)), (
            "map width must be the same in a batch."
        )

        self.num_landmarks = max(
            map(lambda landmark_pos: landmark_pos.shape[0], self.landmark_pos_batch)
        )
        self.free_pos_num = min(map(lambda free_pos: free_pos.shape[0], free_pos_batch))

        if max_free_pos is not None:
            self.free_pos_num = min(
                self.free_pos_num, max_free_pos
            )  # for memory control

        # check free cells is enough for all agents
        assert self.free_pos_num >= self.num_agents, (
            "there is a map without enough number of free cells for agents"
        )

        self.landmark_pos_batch = jnp.stack(
            list(
                map(
                    lambda landmark_pos: pad_placeholder(
                        landmark_pos, self.num_landmarks
                    ),
                    self.landmark_pos_batch,
                )
            ),
            axis=0,
        )

        self.landmark_pos_batch = self.landmark_pos_batch.to_device(ENV_DEVICE)

        free_pos_batch = jnp.stack(
            list(
                map(
                    lambda free_pos: random_truncate(free_pos, self.free_pos_num),
                    free_pos_batch,
                )
            ),
            axis=0,
        )

        free_pos_batch = free_pos_batch.to_device(ENV_DEVICE)

        if agent_idx_batch is not None:
            agent_pos_batch = jax.vmap(idx2pos, in_axes=[0, 0, None, None, None])(
                jnp.array(agent_idx_batch)[:, :, 0],
                jnp.array(agent_idx_batch)[:, :, 1],
                obstacle_size,
                self.height,
                self.width,
            )
            self.generate_agents = lambda key_batch, key_a: jax.random.choice(
                key_batch, agent_pos_batch
            )

        elif random_agents:

            def generate_agents(key_batch, key_a):
                free_pos = jax.random.choice(key_batch, free_pos_batch)
                agent_pos = jax.random.choice(
                    key_a, free_pos, shape=(self.num_agents,), replace=False
                )
                return agent_pos

            self.generate_agents = generate_agents

        else:
            agent_pos_batch = jax.random.choice(
                jax.random.key(0),
                free_pos_batch,
                shape=(self.num_agents,),
                replace=False,
                axis=1,
            )
            self.generate_agents = lambda key_batch, key_a: jax.random.choice(
                key_batch, agent_pos_batch
            )

        if goal_idx_batch is not None:
            goal_pos_batch = jax.vmap(idx2pos, in_axes=[0, 0, None, None, None])(
                jnp.array(goal_idx_batch)[:, :, 0],
                jnp.array(goal_idx_batch)[:, :, 1],
                obstacle_size,
                self.height,
                self.width,
            )
            self.generate_goals = lambda key_batch, key_a: jax.random.choice(
                key_batch, goal_pos_batch
            )

        elif random_goals:

            def generate_goals(key_batch, key_g):
                free_pos = jax.random.choice(key_batch, free_pos_batch)
                goal_pos = jax.random.choice(
                    key_g, free_pos, shape=(self.num_agents,), replace=False
                )
                return goal_pos

            self.generate_goals = generate_goals

        else:
            goal_pos_batch = jax.random.choice(
                jax.random.key(0),
                free_pos_batch,
                shape=(self.num_agents,),
                replace=False,
                axis=1,
            )
            self.generate_goals = lambda key_batch, key_g: jax.random.choice(
                key_batch, goal_pos_batch
            )

    @property
    def landmark_rad(self) -> float:
        return self.obstacle_size / 2

    @property
    def agent_rad(self):
        return self.agent_size / 2

    @property
    def goal_rad(self):
        return self.agent_rad / 2.5

    @property
    def height(self):
        return self._height

    @property
    def width(self):
        return self._width

    def reset(self, key: ArrayLike) -> tuple[Array, Array, Array, Array]:
        key_batch, key = jax.random.split(key, 2)

        key_a, key_g = jax.random.split(key, 2)

        # generate agents
        agent_pos = self.generate_agents(key_batch, key_a)

        # generate goals
        goal_pos = self.generate_goals(key_batch, key_g)

        landmark_pos = jax.random.choice(key_batch, self.landmark_pos_batch)

        return key_g, landmark_pos, agent_pos, goal_pos



import jax
import jax.numpy as jnp

from camar import camar_v0
from camar.environment import Camar
from camar.maps import random_grid, string_grid


class TestEnvironment:
    def test_env_creation(self):
        env = camar_v0()
        assert isinstance(env, Camar)

        map_gen = random_grid(num_agents=4)
        env = camar_v0(map_generator=map_gen)
        assert isinstance(env, Camar)
        assert env.num_agents == 4

    def test_env_reset(self):
        env = camar_v0()
        key = jax.random.PRNGKey(0)
        obs, state = env.reset(key)

        assert obs is not None
        assert state is not None
        assert obs.shape == (env.num_agents, env.observation_size)
        assert state.agent_pos.shape == (env.num_agents, 2)

    def test_env_step(self):
        env = camar_v0(map_generator=random_grid(num_agents=2))
        key = jax.random.PRNGKey(0)
        key_r, key_a, key_s = jax.random.split(key, 3)

        obs, state = env.reset(key_r)
        actions = env.action_spaces.sample(key_a)

        step_fn = jax.jit(env.step)
        obs_next, state_next, reward, done, info = step_fn(key_s, state, actions)

        assert obs_next is not None
        assert state_next is not None
        assert reward is not None
        assert done is not None
        assert info is not None

        assert obs_next.shape == (env.num_agents, env.observation_size)
        assert reward.shape == (env.num_agents, 1)
        assert isinstance(done, jnp.ndarray)

    def test_env_step_with_string_grid(self):
        map_str = """
        .....#.....
        .....#.....
        ...........
        .....#.....
        .....#.....
        """
        env = camar_v0(map_generator=string_grid(map_str=map_str, num_agents=2))
        key = jax.random.PRNGKey(0)
        key_r, key_a, key_s = jax.random.split(key, 3)

        obs, state = env.reset(key_r)
        actions = env.action_spaces.sample(key_a)

        obs_next, state_next, reward, done, info = env.step(key_s, state, actions)

        assert obs_next is not None
        assert state_next is not None
        assert reward is not None
        assert done is not None
        assert info is not None

    def test_batched_env_reset_and_step(self):
        num_envs = 4
        env = camar_v0(map_generator=random_grid(num_agents=2))
        key = jax.random.key(0)
        key_r, key_a, key_s = jax.random.split(key, 3)

        env_reset_fn = jax.jit(jax.vmap(env.reset))
        env_step_fn = jax.jit(jax.vmap(env.step))
        action_sampler = jax.jit(jax.vmap(env.action_spaces.sample))

        keys_r = jax.random.split(key_r, num_envs)
        obs, state = env_reset_fn(keys_r)

        assert obs.shape == (num_envs, env.num_agents, env.observation_size)

        keys_a = jax.random.split(key_a, num_envs)
        actions = action_sampler(keys_a)

        keys_s = jax.random.split(key_s, num_envs)
        obs_next, state_next, reward, done, info = env_step_fn(keys_s, state, actions)

        assert obs_next.shape == (num_envs, env.num_agents, env.observation_size)
        assert reward.shape == (num_envs, env.num_agents, 1)
        assert info is not None



import math
from .const import LANDMARK_COLOR, COLORS
from .utils import hex_to_hsl


class Visualizer:
    def __init__(
        self,
        env,
        state_seq,
        animate_agents=True,
        animate_goals=True,
        animate_landmarks=True,
        fps=None,
        color_step=None,
        use_all_colors=False,
        agent_transparancy=0.8,
    ):
        self.env = env
        self.state_seq = state_seq

        if isinstance(self.state_seq, list) and len(self.state_seq) > 1:
            self.animate_agents = animate_agents
            self.animate_goals = animate_goals
            self.animate_landmarks = animate_landmarks
        else:
            self.animate_agents = False
            self.animate_goals = False
            self.animate_landmarks = False

        if fps is None:
            self.fps = 8 / env.step_dt
        else:
            self.fps = fps

        if isinstance(self.state_seq, list):
            self.keytimes = [
                round(tmp / len(self.state_seq), 8)
                for tmp in range(len(self.state_seq) - 1)
            ]
            self.keytimes.append(1.0)
            self.keytimes = ";".join(map(str, self.keytimes))
            self.dur = round(len(self.state_seq) / self.fps, 3)
        else:
            self.keytimes = None
            self.dur = None

        self.width = self.env.width
        self.height = self.env.height
        self.scale = max(self.width, self.height) / 512

        self.landmark_rad = self.env.landmark_rad
        self.agent_rad = self.env.agent_rad
        self.goal_rad = self.env.goal_rad

        self.use_all_colors = use_all_colors
        self.agent_transparancy = str(agent_transparancy)
        if color_step is not None:
            self.color_step = color_step
        else:
            self.color_step = max(360 // self.env.num_agents, 25)

    def render_landmarks(self):
        if (
            isinstance(self.state_seq, list)
            and len(self.state_seq) > 1
            and self.animate_landmarks
        ):
            landmark_pos = {}
            for state in self.state_seq:
                for landmark, (landmark_x, landmark_y) in enumerate(state.landmark_pos):
                    if landmark not in landmark_pos:
                        landmark_pos[landmark] = {"cx": [], "cy": []}
                    landmark_pos[landmark]["cx"].append(float(landmark_x) / self.scale)
                    landmark_pos[landmark]["cy"].append(float(landmark_y) / self.scale)

            state_seq_svg = []
            for i, landmark in enumerate(landmark_pos):

                hex_color = LANDMARK_COLOR
                hue, saturation, lightness = hex_to_hsl(hex_color)
                color = f"hsl({hue}, {saturation}%, {lightness}%)"

                state_seq_svg.append(f'<circle class="landmark" r="{self.landmark_rad / self.scale}" fill="{color}">')
                for attribute_name in landmark_pos[landmark]:
                    values = ";".join(
                        map(
                            lambda x: f"{x:.3f}", landmark_pos[landmark][attribute_name]
                        )
                    )
                    state_seq_svg.append(
                        f'<animate attributeName="{attribute_name}" dur="{self.dur}s"'
                    )
                    state_seq_svg.append(
                        f'\tkeyTimes="{self.keytimes}" repeatCount="indefinite"'
                    )
                    state_seq_svg.append(f'\tvalues="{values}"/>')

                state_seq_svg.append("</circle>")
                state_seq_svg.append("\n")

            return "\n".join(state_seq_svg)
        else:
            if isinstance(self.state_seq, list):
                state = self.state_seq[0]
            else:
                state = self.state_seq

            landmark_svg = []
            for landmark_x, landmark_y in state.landmark_pos:

                hex_color = LANDMARK_COLOR
                hue, saturation, lightness = hex_to_hsl(hex_color)
                color = f"hsl({hue}, {saturation}%, {lightness}%)"

                landmark_x = float(landmark_x) / self.scale
                landmark_y = float(landmark_y) / self.scale
                landmark_svg.append(
                    f'<circle class="landmark" cx="{landmark_x:.3f}" cy="{landmark_y:.3f}" r="{self.landmark_rad / self.scale}" fill="{color}">  </circle>'
                )

            return "\n".join(landmark_svg)

    def render_goals(self):
        if (
            isinstance(self.state_seq, list)
            and len(self.state_seq) > 1
            and self.animate_goals
        ):
            goal_pos = {}
            for state in self.state_seq:
                for goal, (goal_x, goal_y) in enumerate(state.goal_pos):
                    if goal not in goal_pos:
                        goal_pos[goal] = {"cx": [], "cy": []}
                    goal_pos[goal]["cx"].append(float(goal_x) / self.scale)
                    goal_pos[goal]["cy"].append(float(goal_y) / self.scale)

            state_seq_svg = []
            for i, landmark in enumerate(goal_pos):

                if self.use_all_colors:
                    color = (i * self.color_step) % 360
                    color = f"hsl({color}, 100%, 50%)"
                else:
                    hex_color = COLORS[i % len(COLORS)]
                    hue, saturation, lightness = hex_to_hsl(hex_color)
                    color = f"hsl({hue}, {saturation}%, {lightness}%)"

                state_seq_svg.append(
                    f'<circle class="goal" r="{self.goal_rad / self.scale}" fill="{color}">'
                )
                for attribute_name in goal_pos[landmark]:
                    values = ";".join(
                        map(lambda x: f"{x:.3f}", goal_pos[landmark][attribute_name])
                    )
                    state_seq_svg.append(
                        f'<animate attributeName="{attribute_name}" dur="{self.dur}s"'
                    )
                    state_seq_svg.append(
                        f'\tkeyTimes="{self.keytimes}" repeatCount="indefinite"'
                    )
                    state_seq_svg.append(f'\tvalues="{values}"/>')

                state_seq_svg.append("</circle>")
                state_seq_svg.append("\n")

            return "\n".join(state_seq_svg)
        else:
            if isinstance(self.state_seq, list):
                state = self.state_seq[0]
            else:
                state = self.state_seq

            goal_svg = []
            for i, (goal_x, goal_y) in enumerate(state.goal_pos):

                if self.use_all_colors:
                    color = (i * self.color_step) % 360
                    color = f"hsl({color}, 100%, 50%)"
                else:
                    hex_color = COLORS[i % len(COLORS)]
                    hue, saturation, lightness = hex_to_hsl(hex_color)
                    color = f"hsl({hue}, {saturation}%, {lightness}%)"

                goal_x = float(goal_x) / self.scale
                goal_y = float(goal_y) / self.scale
                goal_svg.append(
                    f'<circle class="goal" cx="{goal_x:.3f}" cy="{goal_y:.3f}" r="{self.goal_rad / self.scale}" fill="{color}">  </circle>'
                )

            return "\n".join(goal_svg)

    def render_agents(self):
        if (
            isinstance(self.state_seq, list)
            and len(self.state_seq) > 1
            and self.animate_agents
        ):
            agent_pos = {}
            for state in self.state_seq:
                for agent, (agent_x, agent_y) in enumerate(state.agent_pos):
                    if agent not in agent_pos:
                        agent_pos[agent] = {"cx": [], "cy": []}

                    agent_pos[agent]["cx"].append(float(agent_x) / self.scale)
                    agent_pos[agent]["cy"].append(float(agent_y) / self.scale)

            state_seq_svg = []
            for i, agent in enumerate(agent_pos):

                if self.use_all_colors:
                    color = (i * self.color_step) % 360
                    color = f"hsl({color}, 100%, 50%)"
                else:
                    hex_color = COLORS[i % len(COLORS)]
                    hue, saturation, lightness = hex_to_hsl(hex_color)
                    color = f"hsla({hue}, {saturation}%, {lightness}%, {self.agent_transparancy})"

                state_seq_svg.append(
                    f'<circle class="agent" r="{self.agent_rad / self.scale}" fill="{color}">'
                )
                for attribute_name in agent_pos[agent]:
                    values = ";".join(
                        map(lambda x: f"{x:.3f}", agent_pos[agent][attribute_name])
                    )
                    state_seq_svg.append(
                        f'<animate attributeName="{attribute_name}" dur="{self.dur}s"'
                    )
                    state_seq_svg.append(
                        f'\tkeyTimes="{self.keytimes}" repeatCount="indefinite"'
                    )
                    state_seq_svg.append(f'\tvalues="{values}"/>')

                state_seq_svg.append("</circle>")
                state_seq_svg.append("\n")

            return "\n".join(state_seq_svg)
        else:
            if isinstance(self.state_seq, list):
                state = self.state_seq[0]
            else:
                state = self.state_seq
            agent_svg = []
            for i, (agent_x, agent_y) in enumerate(state.agent_pos):

                if self.use_all_colors:
                    color = (i * self.color_step) % 360
                    color = f"hsl({color}, 100%, 50%)"
                else:
                    hex_color = COLORS[i % len(COLORS)]
                    hue, saturation, lightness = hex_to_hsl(hex_color)
                    color = f"hsla({hue}, {saturation}%, {lightness}%, {self.agent_transparancy})"

                agent_x = float(agent_x) / self.scale
                agent_y = float(agent_y) / self.scale
                agent_svg.append(
                    f'<circle class="agent" cx="{agent_x:.3f}" cy="{agent_y:.3f}" r="{self.agent_rad / self.scale}" fill="{color}">  </circle>'
                )

            return "\n".join(agent_svg)

    def render(self):
        scaled_width = math.ceil(self.width / self.scale)
        scaled_height = math.ceil(self.height / self.scale)

        view_box = (-self.width / 2 / self.scale,
                    -self.height / 2 / self.scale,
                    self.width / self.scale,
                    self.height / self.scale)

        svg_header = [
            '<?xml version="1.0" encoding="UTF-8"?>',
            '<svg xmlns="http://www.w3.org/2000/svg" xmlns:xlink="http://www.w3.org/1999/xlink"',
            f'\twidth="{scaled_width}" height="{scaled_height}" viewBox="{" ".join(map(str, view_box))}">',
        ]
        svg_header = "\n".join(svg_header)

        definitions = [
            "<style>",
            "\t.landmark {{ }}",
            "\t.agent {{ }}",
            f"\t.goal {{stroke: {LANDMARK_COLOR}; stroke-width: {self.goal_rad / 2 / self.scale};}}",
            "</style>",
        ]
        definitions = "\n".join(definitions)

        svg_header = [svg_header, "\n", "<defs>", definitions, "</defs>"]
        svg_header = "\n".join(svg_header)

        svg_landmark = self.render_landmarks()
        svg_goal = self.render_goals()
        svg_agent = self.render_agents()

        return "\n".join(
            [svg_header, "\n", svg_landmark, "\n", svg_goal, "\n", svg_agent, "</svg>"]
        )

    def save_svg(self, filename="test.svg"):
        with open(filename, "w") as svg_file:
            svg_file.write(self.render())



from typing import Optional

import matplotlib.animation as animation
import matplotlib.pyplot as plt


class Visualizer(object):
	def __init__(
		self,
		env,
		state_seq: list,
		reward_seq=None,
		path=None,
	):
		self.env = env

		self.interval = 50
		self.state_seq = state_seq
		self.reward_seq = reward_seq
		self.path = path
		
		self.init_render()

	def animate(
		self,
		save_fname: Optional[str] = None,
		view: bool = True,
	):
		"""Anim for 2D fct - x (#steps, #pop, 2) & fitness (#steps, #pop)"""
		ani = animation.FuncAnimation(
			self.fig,
			self.update,
			frames=len(self.state_seq),
			blit=False,
			interval=self.interval,
		)
		# Save the animation to a gif
		if save_fname is not None:
			ani.save(save_fname)

		if view:
			plt.show(block=True)

	def init_render(self):
		from matplotlib.patches import Circle, ConnectionPatch
		state = self.state_seq[0]
		
		self.fig, self.ax = plt.subplots(1, 1, figsize=(self.env.width / 4, self.env.height / 4))
		# self.fig, self.ax = plt.subplots(1, 1, figsize=(5, 5))

		if self.path is not None:
			self.ax.plot(self.path[:, 0], self.path[:, 1], lw=1, alpha=0.7)
		
		xlim = (self.env.width + 1) / 2
		ylim = (self.env.height + 1) / 2
		self.ax.set_xlim([- xlim, xlim])
		self.ax.set_ylim([- ylim, ylim])
		
		self.agent_artists = []
		self.con_artists = []
		for i in range(self.env.num_agents):
			# agents
			c = Circle(
				state.agent_pos[i], self.env.agent_rad, color="dodgerblue", alpha=0.8,
			)
			self.ax.add_patch(c)
			self.agent_artists.append(c)

			# goals
			c = Circle(
				state.goal_pos[i], self.env.goal_rad, color="blue", alpha=1,
			)
			self.ax.add_patch(c)

			# connections
			con = ConnectionPatch(
				state.agent_pos[i], state.goal_pos[i], "data", "data", lw=0.3, ls="--", color="dimgray", alpha=0.8,
			)
			self.ax.add_patch(con)
			self.con_artists.append(con)

		
		for i in range(self.env.num_landmarks):
			c = Circle(
				state.landmark_pos[i], self.env.landmark_rad, color="black",
			)
			self.ax.add_patch(c)

		self.ax.set_title(f"Step: {state.step}")

		# self.step_counter = self.ax.text(-1.95, 1.95, f"Step: {state.step}", va="top")
			
	def update(self, frame):
		state = self.state_seq[frame]
		for i, c in enumerate(self.agent_artists):
			c.center = state.agent_pos[i]
		for i, con in enumerate(self.con_artists):
			con.xy1 = state.agent_pos[i]
		
		self.ax.set_title(f"Step: {state.step}")

		# self.step_counter.set_text(f"Step: {state.step}")



from jax import Array
from jax.typing import ArrayLike

import jax
import jax.numpy as jnp
from flax import struct


class Box:
    """
    Minimal jittable class for array-shaped gymnax spaces.
    """

    def __init__(
        self,
        low: float,
        high: float,
        shape: tuple | int,
        dtype: jnp.dtype = jnp.float32,
    ):
        self.low = low
        self.high = high
        self.shape = shape
        self.dtype = dtype

    def sample(self, rng: ArrayLike) -> Array:
        """Sample random action uniformly from 1D continuous range."""
        return jax.random.uniform(
            rng, shape=self.shape, minval=self.low, maxval=self.high, dtype=self.dtype,
        )


@struct.dataclass
class State:
    agent_pos: ArrayLike  # [num_entities, [x, y]]
    agent_vel: ArrayLike  # [n, [x, y]]

    goal_pos: ArrayLike  # [num_agents, [x, y]]
    # obstacle_pos: ArrayLike  # [num_obstacles, [x, y]]
    landmark_pos: ArrayLike  # [num_landmarks, [x, y]]

    # observation: ArrayLike  # [num_agents, max_obs, 2]
    # reward: ArrayLike  # [num_agents]

    is_collision: ArrayLike # [num_agents, ]

    # done: ArrayLike  # bool [num_agents, ]
    step: int  # current step
    # metrics
    on_goal: ArrayLike  # [num_agents, ]
    time_to_reach_goal: ArrayLike  # [num_agents, ]
    num_collisions: ArrayLike # [num_agents, ]
    # flowtime: float # current flowtime
    # makespan: float # current makespan

    goal_keys: ArrayLike  # [num_agents, ] or [] - jax keys for the controllable goal generation (keys are updated only for agents on_goal in lifelong)



from functools import partial

import chex
import jax
import jax.numpy as jnp
from flax import struct


@partial(jax.vmap, in_axes=[0, 0, None, None])
@partial(jax.jit, static_argnames=("effective_rad",))
def check_collision(start_point: chex.Array, end_point: chex.Array, obstacles: chex.Array, effective_rad: float):
    """
    Calculate whether there is a collision between a line segment
    defined by start and end points and a set of circle obstacles.

    :param start_point: The `start_point` parameter represents the starting point of the agent's path.

    :param end_point: The `end_point` parameter represents the end point of the agent's path.

    :param obstacles: The positions of obstacles in the environment with shape
    `(num_obstacles, 2)`.

    :param effective_rad: The effective radius within which a collision is considered to have
    occurred between the agent's path and the obstacles.

    :return: A boolean value indicating whether there is a collision between the agent's path
    and the obstacles.
    """
    # Compute the vector along the agent's path.
    start_to_end = end_point - start_point  # (2, )
    start_to_end_length_sq = jnp.sum(start_to_end * start_to_end)  # ()

    # Compute vectors from start_point to each obstacle center.
    start_to_obs = obstacles - start_point[None, :]  # (num_obstacles, 2)

    # Compute the projection scalar (t) for each obstacle onto the line SE.
    # This gives the position along SE where the perpendicular from the obstacle falls.
    t = (
        jnp.sum(start_to_obs * start_to_end[None, :], axis=1) / start_to_end_length_sq
    )  # (num_obstacles, 2)

    # Clamp t to the interval [0, 1] so that we only consider the segment.
    t = jnp.clip(t, 0.0, 1.0)

    # Compute the closest point on the segment for each obstacle.
    # Broadcasting t to scale the vector SE for each obstacle.
    closest_points = (
        start_point[None, :] + t[:, None] * start_to_end[None, :]
    )  # (num_obstacles, 2)

    # Compute the squared distances from each obstacle center to its closest point on the segment.
    dist = jnp.linalg.norm(closest_points - obstacles, axis=1)

    return jnp.any(dist <= effective_rad)


@struct.dataclass
class RRTState:
    key: chex.PRNGKey
    pos: chex.Array  # (num_samples, num_agents, [x, y])
    parent: chex.Array  # (num_samples, num_agents)
    goal_reached: chex.Array  # (num_agents, )
    idx: int


class RRT:
    def __init__(
        self, env, num_samples, step_size, collision_func=check_collision
    ):
        self.num_samples = num_samples
        self.step_size = step_size

        self.agent_rad = env.agent_rad
        self.num_agents = env.num_agents
        self.goal_rad = env.goal_rad
        self.effective_rad = env.agent_rad + env.landmark_rad
        self.sample_limit = jnp.array(
            [env.width / 2, env.height / 2]
        )

        self.check_collision = collision_func

    @partial(jax.jit, static_argnames=("self",))
    def run(
        self,
        key: chex.PRNGKey,
        start: chex.Array,
        goal: chex.Array,
        landmark_pos: chex.Array,
    ):
        """
        Run RRT algorithm to find a path between start and goal
        start (num_agents, 2)
        goal (num_agents, 2)
        landmark_pos (num_landmarks, 2)
        """

        @jax.jit
        def _condition(state: RRTState):
            return (state.idx < self.num_samples) & ~jnp.all(state.goal_reached)

        @jax.jit
        def _step(state: RRTState):
            """
            :param state: The `state` parameter in the `_step` function represents the current state of the
            RRT.
            :type state: RRTState
            :return: updated `state` variable after performing the RRT step.
            """
            new_key, key_s = jax.random.split(state.key)

            # for agents reached their goals (if previous on goal then stay on it)
            gr_pos = state.pos[state.idx - 1, :, :]  # (num_agents, 2)

            # Sample position, find closest idx and increment towards sampled pos
            sampled_pos = jax.random.uniform(
                key_s,
                (self.num_agents, 2),
                minval=-self.sample_limit,
                maxval=self.sample_limit,
            )

            distances = jnp.linalg.norm(
                state.pos[:, :, :] - sampled_pos[None, :, :], axis=-1
            )  # (rrt_samples, num_agents)
            # closest with previous (valid) samples
            distances = jnp.where(
                (jnp.arange(self.num_samples) < state.idx)[:, None], distances, jnp.inf
            )
            distances = jnp.where(state.parent == -2.0, jnp.inf, distances)
            closest_idx = jnp.argmin(distances, axis=0)  # (num_agents, )

            tree_pos = state.pos[
                closest_idx, jnp.arange(self.num_agents), :
            ]  # (num_agents, 2)

            direction = sampled_pos - tree_pos
            direction_dist = distances[
                closest_idx, jnp.arange(self.num_agents)
            ]  # (num_agents, )
            clipped_direction = jnp.where(
                (direction_dist > self.step_size)[:, None],
                direction / direction_dist[:, None] * self.step_size,
                direction,
            )  # (num_agents, 2)
            test_pos = tree_pos + clipped_direction

            # Check free space, line collision
            valid = ~self.check_collision(
                tree_pos, test_pos, landmark_pos, self.effective_rad
            )  # (num_agents, )

            new_pos = jnp.where(valid[:, None], test_pos, -2.0)
            new_parent = jnp.where(valid, closest_idx, -2)

            new_pos = jnp.where(state.goal_reached[:, None], gr_pos, new_pos)
            new_parent = jnp.where(state.goal_reached, -1, new_parent)

            gr_test = jnp.linalg.norm(test_pos - goal, axis=-1) < self.goal_rad
            gr_test = jnp.where(valid, gr_test, False)
            goal_reached = jnp.where(state.goal_reached, True, gr_test)

            state = state.replace(
                key=new_key,
                pos=state.pos.at[state.idx].set(new_pos),
                parent=state.parent.at[state.idx].set(new_parent),
                goal_reached=goal_reached,
                idx=state.idx + 1 * jnp.any(valid),
            )

            return state

        state = RRTState(
            key=key,
            pos=jnp.full(
                (self.num_samples, self.num_agents, 2), start, dtype=jnp.float32
            ),
            parent=jnp.full((self.num_samples, self.num_agents), -1, dtype=jnp.int32),
            goal_reached=jnp.linalg.norm(start - goal, axis=-1) < self.goal_rad,
            idx=1,
        )

        res_state = jax.lax.while_loop(_condition, _step, state)
        return res_state

    @partial(jax.jit, static_argnames=["self"])
    def find_last_idx(self, state: RRTState):
        idx = jnp.arange(self.num_samples)
        valid_idx = jnp.where(
            (state.parent != -1) & (state.parent != -2), idx[:, None], -1
        )
        return jnp.max(valid_idx, axis=0)



from functools import partial
from typing import List, Optional

import jax.numpy as jnp

from .base_map import base_map
from .batched_string_grid import batched_string_grid
from .const import PREGEN_DEVICE
from .utils import detect_edges, get_movingai


def preprocess(map_array, height, width, low_thr, return_edges=True):
    import cv2
    import numpy as np

    map_array = np.array(map_array)
    map_array = cv2.resize(map_array, (height, width), interpolation=cv2.INTER_NEAREST)
    map_array = jnp.asarray(map_array, device=PREGEN_DEVICE)

    if return_edges:
        map_array = detect_edges(map_array, low_thr)
    return map_array


class movingai(batched_string_grid):
    def __init__(
        self,
        map_names: List[str],
        height: int = 128,
        width: int = 128,
        low_thr: float = 3.7,
        only_edges: bool = True,
        remove_border: bool = True,
        add_border: bool = False,
        num_agents: int = 10,
        obstacle_size: float = 0.2,
        agent_size: float = 0.1,
        max_free_pos: Optional[int] = None,
    ) -> base_map:
        map_str_batch = get_movingai(map_names)

        map_array_preprocess = partial(
            preprocess,
            height=height,
            width=width,
            low_thr=low_thr,
            return_edges=only_edges,
        )
        free_pos_array_preprocess = partial(
            preprocess,
            height=height,
            width=width,
            low_thr=low_thr,
            return_edges=False,
        )

        super().__init__(
            map_str_batch=map_str_batch,
            free_pos_str_batch=map_str_batch,
            agent_idx_batch=None,
            goal_idx_batch=None,
            num_agents=num_agents,
            random_agents=True,
            random_goals=True,
            remove_border=remove_border,
            add_border=add_border,
            obstacle_size=obstacle_size,
            agent_size=agent_size,
            map_array_preprocess=map_array_preprocess,
            free_pos_array_preprocess=free_pos_array_preprocess,
            max_free_pos=max_free_pos,
        )



import os

os.environ['CUDA_VISIBLE_DEVICES'] = '0'
os.environ['XLA_PYTHON_CLIENT_PREALLOCATE'] = 'false'
os.environ['PYTORCH_CUDA_ALLOC_CONF'] = 'expandable_segments:True'

import dataclasses

import matplotlib.pyplot as plt
import pandas as pd
import torch
from tensordict.nn import TensorDictModule, TensorDictSequential
from tensordict.nn.distributions import NormalParamExtractor
from torchrl.collectors import SyncDataCollector
from torchrl.data.replay_buffers import RandomSampler, ReplayBuffer
from torchrl.data.replay_buffers.storages import LazyTensorStorage
from torchrl.envs import RewardSum, TransformedEnv
from torchrl.modules import MultiAgentMLP, ProbabilisticActor, TanhNormal
from torchrl.objectives import SACLoss, SoftUpdate, ValueEstimators
from tqdm.auto import tqdm

import wandb
from camar import camar_v0
from camar.maps import random_grid
from camar.render import SVG_Visualizer
from camar.integrations.torchrl import CamarWrapper

os.chdir(os.getcwd() + "/baselines/")
torch.manual_seed(911)

model_name = "ISAC"

device = torch.device("cuda:0")

frames_per_batch = 54_000
n_iters = 150
total_frames = frames_per_batch * n_iters

# Replay buffer
memory_size = 1_000_000

n_optimiser_steps = 100
train_batch_size = 5000
lr = 3e-4
max_grad_norm = 1.0

gamma = 0.99
polyak_tau = 0.005



num_rows = 10
num_cols = 10
obstacle_density = 0.0
num_agents = 5
grain_factor = 6

window = 0.8

map_generator = random_grid(
    num_rows = num_rows,
    num_cols = num_cols,
    obstacle_density = obstacle_density,
    num_agents = num_agents,
    grain_factor = grain_factor,
)

env = camar_v0(
    map_generator = map_generator,
    window = window,
    lifelong = False,
    max_steps = 900,
    contact_force=500,
    contact_margin=1e-3,
    dt=0.01,
    frameskip=7,
    max_obs=8,
)

num_envs = frames_per_batch // env.max_steps
print("num_envs =", num_envs)

env = CamarWrapper(env=env, device=device, batch_size=[num_envs], seed=0)

env = TransformedEnv(
    env,
    RewardSum(in_keys=[env.reward_key], out_keys=[("agents", "episode_reward")]),
)

# set up policy

policy_net = torch.nn.Sequential(
    MultiAgentMLP(
        n_agent_inputs=env.observation_spec["agents", "observation"].shape[-1],
        n_agent_outputs=2 * env.action_spec.shape[-1],
        n_agents=num_agents,
        device=device,
        depth=2,
        num_cells=256,
        activation_class=torch.nn.Tanh,
        share_params=True,
        centralized=False,
        # layer_class=NoisyLinear, #doesn't work
    ),
    NormalParamExtractor(),
)

policy_module = TensorDictModule(
    policy_net,
    in_keys=[("agents", "observation")],
    out_keys=[("agents", "loc"), ("agents", "scale")],
)

policy = ProbabilisticActor(
    module=policy_module,
    spec=env.full_action_spec["agents", "action"],
    in_keys=[("agents", "loc"), ("agents", "scale")],
    out_keys=[env.action_key],
    distribution_class=TanhNormal,
    distribution_kwargs={
        "low": env.action_spec.space.low,
        "high": env.action_spec.space.high,
    },
    return_log_prob=True,
    log_prob_key=("agents", "log_prob"),
)

# set up q network

cat_module = TensorDictModule(
    lambda obs, action: torch.cat([obs, action], dim=-1),
    in_keys=[("agents", "observation"), ("agents", "action")],
    out_keys=[("agents", "obs_action")],
)

qvalue_net = MultiAgentMLP(
    n_agent_inputs=env.observation_spec["agents", "observation"].shape[-1] + env.full_action_spec["agents", "action"].shape[-1],
    n_agent_outputs=1,
    n_agents=num_agents,
    device=device,
    depth=2,
    num_cells=256,
    activation_class=torch.nn.Tanh,
    share_params=True, # can be changed
    centralized=False, # True for masac and False for isac
)

qvalue_module = TensorDictModule(
    module=qvalue_net,
    in_keys=[("agents", "obs_action")],
    out_keys=[("agents", "state_action_value")],
)

qvalue = TensorDictSequential(
    cat_module, qvalue_module,
)

collector = SyncDataCollector(
    env,
    policy,
    device=device,
    storing_device=device,
    frames_per_batch=frames_per_batch,
    total_frames=total_frames,
)

replay_buffer = ReplayBuffer(
    storage=LazyTensorStorage(
        memory_size, device=device
    ),
    sampler=RandomSampler(),
    batch_size=train_batch_size,
)

loss_module = SACLoss(
    actor_network=policy,
    qvalue_network=qvalue,
    # value_network=value,
)

loss_module.set_keys(
    reward=env.reward_key,
    action=env.action_key,
    log_prob=("agents", "log_prob"),
    state_action_value=("agents", "state_action_value"),
    value=("agents", "state_value"),
    done=("agents", "done"),
    terminated=("agents", "terminated"),
)
loss_module.make_value_estimator(ValueEstimators.TD0, gamma=gamma)

target_updater = SoftUpdate(loss_module, tau=polyak_tau)

optim = torch.optim.Adam(loss_module.parameters(), lr)

def process_batch(batch):
    """
    If the `(group, "terminated")` and `(group, "done")` keys are not present, create them by expanding
    `"terminated"` and `"done"`.
    This is needed to present them with the same shape as the reward to the loss.
    """
    keys = list(batch.keys(True, True))
    group_shape = batch.get_item_shape("agents")
    nested_done_key = ("next", "agents", "done")
    nested_terminated_key = ("next", "agents", "terminated")
    if nested_done_key not in keys:
        batch.set(
            nested_done_key,
            batch.get(("next", "done")).unsqueeze(-1).expand((*group_shape, 1)),
        )
    if nested_terminated_key not in keys:
        batch.set(
            nested_terminated_key,
            batch.get(("next", "terminated"))
            .unsqueeze(-1)
            .expand((*group_shape, 1)),
        )
    return batch

pbar = tqdm(total=n_iters, desc="episode_reward_mean = 0")

wandb.init(
    project="myenv_xppo",
    config={
        "num_rows": num_rows,
        "num_cols": num_cols,
        "obstacle_density": obstacle_density,
        "num_agents": num_agents,
        "grain_factor": grain_factor,
        "frames_per_batch": frames_per_batch,
        "n_iters": n_iters,
        "total_frames": total_frames,
        "n_optimiser_steps": n_optimiser_steps,
        "train_batch_size": train_batch_size,
        "lr": lr,
        "max_grad_norm": max_grad_norm,
        "gamma": gamma,
        "polyak_tau": polyak_tau,
    },
    name=f"{model_name}"
)

plot_data = {
    "env_steps": [],
    "ep_rew_mean": [],
}

env_steps = 0
for iteration, batch in enumerate(collector):
    current_frames = batch.numel()
    batch = process_batch(batch)  # Util to expand done keys if needed

    data_view = batch.reshape(-1) # This just affects the leading dimensions in batch_size of the tensordict
    replay_buffer.extend(data_view)

    for _ in range(n_optimiser_steps):
        subdata = replay_buffer.sample()
        loss_vals = loss_module(subdata)

        loss_total = (
                loss_vals["loss_actor"]
                + loss_vals["loss_alpha"]
                + loss_vals["loss_qvalue"]
            )

        grad_norms = {}

        loss_total.backward()

        grad_norm = torch.nn.utils.clip_grad_norm_(
            loss_module.parameters(), max_grad_norm
        )

        wandb.log({
                    "loss_actor": loss_vals["loss_actor"].item(),
                    "loss_alpha": loss_vals["loss_alpha"].item(),
                    "loss_qvalue": loss_vals["loss_qvalue"].item(),
                    "loss_total": loss_total.item(),
                    "grad_norm": grad_norm,
        })

        optim.step()
        optim.zero_grad()

        target_updater.step()

    done = batch.get(("next", "done"))[:, :, 0]
    episode_reward_mean = batch.get(("next", "agents", "episode_reward"))[done].mean().item()

    env_steps += data_view.shape[0]

    plot_data["env_steps"].append(env_steps)
    plot_data["ep_rew_mean"].append(episode_reward_mean)

    wandb.log({
        "ep_rew_mean": episode_reward_mean,
        "iter": iteration,
        "env_steps": env_steps,
    })

    pbar.set_description(f"episode_reward_mean = {episode_reward_mean :.2f}", refresh=False)
    pbar.update()

wandb.finish()


plot_df = pd.DataFrame(plot_data)

fig = plt.figure()
ax = fig.add_subplot()

ax.plot(plot_df["env_steps"], plot_df["ep_rew_mean"])
ax.set_xlabel("env_steps")
ax.set_ylabel("ep_rew_mean")

fig.savefig(f"img/ep_rew_mean_{model_name}.png")
plot_df.to_csv(f"data/ep_rew_mean_{model_name}.csv")


def get_state_from_envs(state, env_id):
    state_data = {
        field.name: getattr(state, field.name)[env_id]
        for field in dataclasses.fields(state)
    }
    return type(state)(**state_data)


def rendering_callback(env, td):
    env.state_seq.append(get_state_from_envs(env._state, 0))


viz_env = camar_v0(
    map_generator = map_generator,
    window = window,
    lifelong = False,
    max_steps = 900,
    contact_force=500,
    contact_margin=1e-3,
    dt=0.01,
    frameskip=7,
    max_obs=8,
)

viz_env = CamarWrapper(env=viz_env, device=device, batch_size=[2], seed=5)

viz_env = TransformedEnv(
    viz_env,
    RewardSum(in_keys=[viz_env.reward_key], out_keys=[("agents", "episode_reward")]),
)


viz_env.state_seq = []

with torch.no_grad():
    out = viz_env.rollout(
        auto_reset=True,
        max_steps=viz_env.max_steps + 1,
        policy=policy,
        callback=rendering_callback,
        auto_cast_to_device=True,
        break_when_any_done=True,
    )


SVG_Visualizer(viz_env._env, viz_env.state_seq).save_svg(f"env_viz/example_{model_name}.svg")



# From Craftax Baselines https://github.com/MichaelTMatthews/Craftax_Baselines.git with added a success rate metric

from functools import partial
from typing import Any, Union

import jax
from flax import struct
from jax.typing import ArrayLike

from camar.wrappers import GymnaxWrapper


@struct.dataclass
class LogEnvState:
    env_state: Any
    episode_returns: float
    episode_lengths: int
    returned_episode_returns: float
    returned_episode_lengths: int
    timestep: int
    success_rate: float

class LogWrapper(GymnaxWrapper):
    """Log the episoreturns and lengths."""

    def __init__(self, env):
        super().__init__(env)


    @partial(jax.jit, static_argnums=(0, ))
    def reset(self, key: ArrayLike):
        obs, env_state = self._env.reset(key)
        state = LogEnvState(env_state, 0.0, 0, 0.0, 0, 0)
        return obs, state

    @partial(jax.jit, static_argnums=(0, ))
    def step(
        self,
        key: ArrayLike,
        state,
        action: Union[int, float],
    ):
        obs, env_state, reward, done, info = self._env.step(key, state.env_state, action)
        new_episode_return = state.episode_returns + reward
        new_episode_length = state.episode_lengths + 1
        state = LogEnvState(
            env_state=env_state,
            episode_returns=new_episode_return * (1 - done),
            episode_lengths=new_episode_length * (1 - done),
            returned_episode_returns=state.returned_episode_returns * (1 - done)
            + new_episode_return * done,
            returned_episode_lengths=state.returned_episode_lengths * (1 - done)
            + new_episode_length * done,
            timestep=state.timestep + 1,
            success_rate=env_state.on_goal.sum(axis=-1) / self.num_agents,
        )
        info["returned_episode_returns"] = state.returned_episode_returns
        info["returned_episode_lengths"] = state.returned_episode_lengths
        info["timestep"] = state.timestep
        info["returned_episode"] = done

        info["success_rate"] = state.success_rate

        return obs, state, reward, done, info



import importlib
from typing import Optional, Union

from .environment import Camar
from .maps import base_map

MAPS_MODULE = "camar.maps"


def make_env(
    map_generator: Optional[Union[str, base_map]] = "random_grid",
    window: Optional[float] = None,
    placeholder: float = 0.0,
    max_steps: int = 100,
    frameskip: int = 2,
    max_obs: Optional[int] = None,
    dt: float = 0.01,
    damping: float = 0.25,
    contact_force: float = 500,
    contact_margin: float = 0.001,
    **map_kwargs,
):
    if isinstance(map_generator, str):
        module = importlib.import_module(MAPS_MODULE)
        map_generator = getattr(module, map_generator)(**map_kwargs)

    env = Camar(
        map_generator=map_generator,
        window=window,
        placeholder=placeholder,
        max_steps=max_steps,
        frameskip=frameskip,
        max_obs=max_obs,
        dt=dt,
        damping=damping,
        contact_force=contact_force,
        contact_margin=contact_margin,
    )

    return env



import jax
import pytest

from camar.maps import random_grid


class TestRandomGrid:
    def test_map_creation(self):
        map_gen = random_grid()
        assert map_gen is not None
        assert map_gen.num_agents > 0
        assert map_gen.num_landmarks > 0

    def test_map_reset(self):
        map_gen = random_grid(num_agents=4, num_rows=10, num_cols=10)
        key = jax.random.key(0)
        key_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert key_g is not None
        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape == (map_gen.num_landmarks, 2)
        assert agent_pos.shape == (map_gen.num_agents, 2)
        assert goal_pos.shape == (map_gen.num_agents, 2)

    def test_map_properties(self):
        map_gen = random_grid(obstacle_size=0.5, grain_factor=3)
        assert map_gen.landmark_rad == pytest.approx(0.5 / (2 * (3 - 1)))



import jax
import pytest

from camar.maps import movingai

TEST_MAP_NAMES = ["street/Denver_0_1024"]


class TestMovingAI:
    def test_map_creation(self):
        # try:
        map_gen = movingai(map_names=TEST_MAP_NAMES, num_agents=2)
        assert map_gen is not None
        assert map_gen.num_agents == 2
        assert map_gen.batch_size == len(TEST_MAP_NAMES)
        assert map_gen.num_landmarks > 0

    # except Exception as e:
    #     pytest.skip(
    #         f"Skipping MovingAI test due to potential network/data issue: {e}"
    #     )

    def test_map_reset(self):
        # try:
        num_agents = 2
        map_gen = movingai(map_names=TEST_MAP_NAMES, num_agents=num_agents)
        key = jax.random.key(0)

        keys_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape == (map_gen.num_landmarks, 2)
        assert agent_pos.shape == (num_agents, 2)
        assert goal_pos.shape == (num_agents, 2)

    # except Exception as e:
    #     pytest.skip(
    #         f"Skipping MovingAI test due to potential network/data issue: {e}"
    #     )

    def test_map_properties(self):
        # try:
        obstacle_size = 0.15
        agent_size = 0.07
        map_gen = movingai(
            map_names=TEST_MAP_NAMES,
            obstacle_size=obstacle_size,
            agent_size=agent_size,
        )

        assert map_gen.landmark_rad == pytest.approx(obstacle_size / 2)
        assert map_gen.agent_rad == pytest.approx(agent_size / 2)
        assert map_gen.goal_rad == pytest.approx((agent_size / 2) / 2.5)

        resized_h, resized_w = 128, 128

        assert map_gen.height == pytest.approx((resized_h - 2) * obstacle_size)
        assert map_gen.width == pytest.approx((resized_w - 2) * obstacle_size)

    # except Exception as e:
    #     pytest.skip(
    #         f"Skipping MovingAI test due to potential network/data issue: {e}"
    #     )



import jax
import pytest

from camar.maps import labmaze_grid


class TestLabMazeGrid:
    def test_map_creation(self):
        num_maps = 2
        map_gen = labmaze_grid(num_maps=num_maps, num_agents=4, height=7, width=7)
        assert map_gen is not None
        assert map_gen.num_agents == 4
        assert map_gen.batch_size == num_maps
        assert map_gen.num_landmarks > 0

    def test_map_reset(self):
        num_maps = 2
        num_agents = 3
        map_gen = labmaze_grid(
            num_maps=num_maps,
            num_agents=num_agents,
            height=7,
            width=7,
            room_min_size=2,
            room_max_size=3,
        )
        key = jax.random.key(0)

        keys_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert keys_g is not None
        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape == (map_gen.num_landmarks, 2)
        assert agent_pos.shape == (num_agents, 2)
        assert goal_pos.shape == (num_agents, 2)

    def test_map_properties(self):
        obstacle_size = 0.25
        agent_size = 0.12
        map_gen = labmaze_grid(
            num_maps=1,
            obstacle_size=obstacle_size,
            agent_size=agent_size,
            height=7,
            width=7,
            room_min_size=2,
            room_max_size=3,
        )

        assert map_gen.landmark_rad == pytest.approx(obstacle_size / 2)
        assert map_gen.agent_rad == pytest.approx(agent_size / 2)
        assert map_gen.goal_rad == pytest.approx((agent_size / 2) / 2.5)

        init_height, init_width = 7, 7
        map_gen_dims = labmaze_grid(
            num_maps=1,
            height=init_height,
            width=init_width,
            obstacle_size=obstacle_size,
        )
        assert map_gen_dims.height == pytest.approx(init_height * obstacle_size)
        assert map_gen_dims.width == pytest.approx(init_width * obstacle_size)



import jax
import pytest

from camar.maps import caves_cont


class TestCavesCont:
    def test_map_creation(self):
        map_gen = caves_cont()
        assert map_gen is not None
        assert map_gen.num_agents > 0
        assert map_gen.num_landmarks > 0

        map_gen_no_border = caves_cont(add_borders=False)
        assert map_gen_no_border is not None
        assert map_gen_no_border.border_landmarks.shape[0] == 0

    def test_map_reset(self):
        num_agents = 4
        map_gen = caves_cont(
            num_agents=num_agents, num_rows=32, num_cols=32, scale=7, free_ratio=0.5
        )
        key = jax.random.key(0)
        key_g, landmark_pos, agent_pos, goal_pos = map_gen.reset(key)

        assert key_g is not None
        assert landmark_pos is not None
        assert agent_pos is not None
        assert goal_pos is not None

        assert landmark_pos.shape == (map_gen.num_landmarks, 2)
        assert agent_pos.shape == (num_agents, 2)
        assert goal_pos.shape == (num_agents, 2)

    def test_map_properties(self):
        obstacle_size = 0.2
        agent_size = 0.1
        map_gen = caves_cont(obstacle_size=obstacle_size, agent_size=agent_size)
        assert map_gen.landmark_rad == pytest.approx(obstacle_size / 2)
        assert map_gen.agent_rad == pytest.approx(agent_size / 2)
        assert map_gen.goal_rad == pytest.approx((agent_size / 2) / 2.5)
        assert map_gen.height == map_gen.num_cols * obstacle_size
        assert map_gen.width == map_gen.num_rows * obstacle_size

    def test_invalid_landmark_ratios(self):
        with pytest.raises(ValueError):
            caves_cont(landmark_low_ratio=0.8, landmark_high_ratio=0.7)



import os

os.environ["CUDA_VISIBLE_DEVICES"] = "0"
os.environ["XLA_PYTHON_CLIENT_PREALLOCATE"] = "false"
os.environ["PYTORCH_CUDA_ALLOC_CONF"] = "expandable_segments:True"

import dataclasses

import matplotlib.pyplot as plt
import pandas as pd
import torch
from tensordict.nn import TensorDictModule
from tensordict.nn.distributions import NormalParamExtractor
from torchrl.collectors import SyncDataCollector
from torchrl.data.replay_buffers import ReplayBuffer
from torchrl.data.replay_buffers.samplers import SamplerWithoutReplacement
from torchrl.data.replay_buffers.storages import LazyTensorStorage
from torchrl.envs import RewardSum, TransformedEnv
from torchrl.modules import MultiAgentMLP, ProbabilisticActor, TanhNormal
from torchrl.objectives import ClipPPOLoss, ValueEstimators
from tqdm.auto import tqdm

import wandb
from camar import camar_v0
from camar.maps import random_grid
from camar.render import SVG_Visualizer
from camar.integrations.torchrl import CamarWrapper


os.chdir(os.getcwd() + "/baselines/")
torch.manual_seed(911)

# parameters
params = {
    "CPPO": {
        "policy": {"share_params": True, "centralised": True},
        "critic": {"share_params": True, "centralised": True},
    },
    "MAPPO": {
        "policy": {"share_params": True, "centralised": False},
        "critic": {"share_params": True, "centralised": True},
    },
    "IPPO": {
        "policy": {"share_params": True, "centralised": False},
        "critic": {"share_params": True, "centralised": False},
    },
    "HetIPPO": {
        "policy": {"share_params": False, "centralised": False},
        "critic": {"share_params": False, "centralised": False},
    },
}

model_name = "MAPPO"

policy_params = params[model_name]["policy"]
critic_params = params[model_name]["critic"]

device = torch.device("cuda:0")

frames_per_batch = 200_000
n_iters = 40
total_frames = frames_per_batch * n_iters

num_epochs = 2
minibatch_size = 500
lr = 3e-4
max_grad_norm = 1.0

clip_epsilon = 0.2
gamma = 0.99
lmbda = 0.9
entropy_eps = 1e-4
critic_coef = 1
loss_critic_type = "smooth_l2"

wandb_name = "MAPPO_0.5_on_goal_1.0_collision_1.0_rew_diff"

num_rows = 10
num_cols = 10
obstacle_density = 0.0
num_agents = 5
grain_factor = 6

window = 0.8

map_generator = random_grid(
    num_rows = num_rows,
    num_cols = num_cols,
    obstacle_density = obstacle_density,
    num_agents = num_agents,
    grain_factor = grain_factor,
)

env = camar_v0(
    map_generator = map_generator,
    window = window,
    lifelong = False,
    max_steps = 1000,
    contact_force=500,
    contact_margin=1e-3,
    dt=0.01,
    frameskip=7,
    max_obs=8,
)

num_envs = frames_per_batch // env.max_steps
print("num_envs =", num_envs)

env = CamarWrapper(env=env, device=device, batch_size=[num_envs], seed=0)

env = TransformedEnv(
    env,
    RewardSum(in_keys=[env.reward_key], out_keys=[("agents", "episode_reward")]),
)

policy_net = torch.nn.Sequential(
    MultiAgentMLP(
        n_agent_inputs=env.observation_spec["agents", "observation"].shape[-1],
        n_agent_outputs=2 * env.action_spec.shape[-1],
        n_agents=num_agents,
        device=device,
        depth=2,
        num_cells=256,
        activation_class=torch.nn.Tanh,
        share_params=policy_params["share_params"],
        centralized=policy_params["centralised"],
        # layer_class=NoisyLinear, #doesn't work
    ),
    NormalParamExtractor(),
)

policy_module = TensorDictModule(
    policy_net,
    in_keys=[("agents", "observation")],
    out_keys=[("agents", "loc"), ("agents", "scale")],
)

policy = ProbabilisticActor(
    module=policy_module,
    # spec=env.unbatched_action_spec,
    in_keys=[("agents", "loc"), ("agents", "scale")],
    out_keys=[env.action_key],
    distribution_class=TanhNormal,
    distribution_kwargs={
        "low": env.action_spec.space.low,
        "high": env.action_spec.space.high,
    },
    return_log_prob=True,
    log_prob_key=("agents", "sample_log_prob"),
)

critic_net = MultiAgentMLP(
    n_agent_inputs=env.observation_spec["agents", "observation"].shape[-1],
    n_agent_outputs=1,
    n_agents=num_agents,
    device=device,
    depth=2,
    num_cells=256,
    activation_class=torch.nn.Tanh,
    share_params=critic_params["share_params"],
    centralized=critic_params["centralised"],
)

critic = TensorDictModule(
    module=critic_net,
    in_keys=[("agents", "observation")],
    out_keys=[("agents", "state_value")],
)

collector = SyncDataCollector(
    env,
    policy,
    device=device,
    storing_device=device,
    frames_per_batch=frames_per_batch,
    total_frames=total_frames,
)

replay_buffer = ReplayBuffer(
    storage=LazyTensorStorage(frames_per_batch, device=device),
    sampler=SamplerWithoutReplacement(),
    batch_size=minibatch_size,
)

loss_module = ClipPPOLoss(
    actor_network=policy,
    critic_network=critic,
    critic_coef=critic_coef,
    clip_epsilon=clip_epsilon,
    entropy_coef=entropy_eps,
    normalize_advantage=False,
    entropy_bonus=True,
)
loss_module.set_keys(
    reward=env.reward_key,
    action=env.action_key,
    sample_log_prob=("agents", "sample_log_prob"),
    value=("agents", "state_value"),
    done=("agents", "done"),
    terminated=("agents", "terminated"),
)


loss_module.make_value_estimator(ValueEstimators.GAE, gamma=gamma, lmbda=lmbda)
GAE = loss_module.value_estimator

optim = torch.optim.Adam(loss_module.parameters(), lr)

pbar = tqdm(total=n_iters, desc="episode_reward_mean = 0")

wandb.init(
    project="myenv_xppo",
    config={
        "num_rows": num_rows,
        "num_cols": num_cols,
        "obstacle_density": obstacle_density,
        "num_agents": num_agents,
        "grain_factor": grain_factor,
        "frames_per_batch": frames_per_batch,
        "n_iters": n_iters,
        "total_frames": total_frames,
        "num_epochs": num_epochs,
        "minibatch_size": minibatch_size,
        "lr": lr,
        "max_grad_norm": max_grad_norm,
        "clip_epsilon": clip_epsilon,
        "gamma": gamma,
        "lmbda": lmbda,
        "entropy_eps": entropy_eps,
        "critic_coef": critic_coef,
        "loss_critic_type": loss_critic_type,
    },
    name=wandb_name,
)

plot_data = {
    "env_steps": [],
    "ep_rew_mean": [],
}

env_steps = 0
for iter, tensordict_data in enumerate(collector):
    tensordict_data.set(
        ("next", "agents", "done"),
        tensordict_data.get(("next", "done"))
        .unsqueeze(-1)
        .expand(tensordict_data.get_item_shape(("next", env.reward_key))),
    )
    tensordict_data.set(
        ("next", "agents", "terminated"),
        tensordict_data.get(("next", "terminated"))
        .unsqueeze(-1)
        .expand(tensordict_data.get_item_shape(("next", env.reward_key))),
    )

    with torch.no_grad():
        GAE(
            tensordict_data,
            params=loss_module.critic_network_params,
            target_params=loss_module.target_critic_network_params,
        )

    data_view = tensordict_data.reshape(-1)
    replay_buffer.extend(data_view)

    for num_epoch in range(num_epochs):
        for num_batch in range(frames_per_batch // minibatch_size):
            num_grad_step = (
                iter * num_epochs * (frames_per_batch // minibatch_size)
                + num_epoch * (frames_per_batch // minibatch_size)
                + num_batch
            )

            subdata = replay_buffer.sample()
            loss_vals = loss_module(subdata)

            loss_value = (
                loss_vals["loss_objective"]
                + loss_vals["loss_critic"]
                + loss_vals["loss_entropy"]
            )

            loss_value.backward()

            grad_norm = torch.nn.utils.clip_grad_norm_(
                loss_module.parameters(), max_grad_norm
            )

            wandb.log(
                {
                    "mean_adv": subdata["advantage"].mean().item(),
                    "loss_obj": loss_vals["loss_objective"].item(),
                    "loss_critic": loss_vals["loss_critic"].item(),
                    "loss_entropy": loss_vals["loss_entropy"].item(),
                    "loss_total": loss_value.item(),
                    "grad_norm": grad_norm,
                    "num_grad_step": num_grad_step,
                }
            )

            optim.step()
            optim.zero_grad()

    collector.update_policy_weights_()

    done = tensordict_data.get(("next", "done"))[:, :, 0]
    episode_reward_mean = (
        tensordict_data.get(("next", "agents", "episode_reward"))[done].mean().item()
    )
    # episode_reward_mean = (
    # 	tensordict_data.get(("next", "agents", "episode_reward")).mean().item()
    # )

    env_steps += data_view.shape[0]

    plot_data["env_steps"].append(env_steps)
    plot_data["ep_rew_mean"].append(episode_reward_mean)

    wandb.log(
        {
            "ep_rew_mean": episode_reward_mean,
            "iter": iter,
            "env_steps": env_steps,
        }
    )

    pbar.set_description(
        f"episode_reward_mean = {episode_reward_mean:.2f}", refresh=False
    )
    pbar.update()

wandb.finish()

plot_df = pd.DataFrame(plot_data)

fig = plt.figure()
ax = fig.add_subplot()

ax.plot(plot_df["env_steps"], plot_df["ep_rew_mean"])
ax.set_xlabel("env_steps")
ax.set_ylabel("ep_rew_mean")

fig.savefig(f"img/ep_rew_mean_{model_name}.png")
plot_df.to_csv(f"data/ep_rew_mean_{model_name}.csv")


def get_state_from_envs(state, env_id):
    state_data = {
        field.name: getattr(state, field.name)[env_id]
        for field in dataclasses.fields(state)
    }
    return type(state)(**state_data)


def rendering_callback(env, td):
    env.state_seq.append(get_state_from_envs(env._state, 0))


viz_env = camar_v0(
    map_generator = map_generator,
    window = window,
    lifelong = False,
    max_steps = 900,
    contact_force=500,
    contact_margin=1e-3,
    dt=0.01,
    frameskip=7,
    max_obs=8,
)

viz_env = CamarWrapper(env=viz_env, device=device, batch_size=[2], seed=5)

viz_env = TransformedEnv(
    viz_env,
    RewardSum(in_keys=[viz_env.reward_key], out_keys=[("agents", "episode_reward")]),
)


viz_env.state_seq = []

with torch.no_grad():
    out = viz_env.rollout(
        auto_reset=True,
        max_steps=viz_env.max_steps + 1,
        policy=policy,
        callback=rendering_callback,
        auto_cast_to_device=True,
        break_when_any_done=True,
    )


SVG_Visualizer(viz_env._env, viz_env.state_seq).save_svg(f"env_viz/example_{model_name}.svg")



from typing import Optional

from .base_map import base_map
from .batched_string_grid import batched_string_grid


def generate_labmaze_maps(num_maps, height, width, max_rooms, seed, **labmaze_kwargs):
    from labmaze import RandomMaze

    maps = []
    free_pos = []
    for i in range(num_maps):
        random_maze = RandomMaze(
            height=height,
            width=width,
            max_rooms=max_rooms,
            random_seed=seed + i,
            **labmaze_kwargs,
        )
        maps.append(str(random_maze.entity_layer))
        free_pos.append(str(random_maze.variations_layer).replace(".", "*"))
    return maps, free_pos


class labmaze_grid(batched_string_grid):
    def __init__(
        self,
        num_maps: int,
        height: int = 11,
        width: int = 11,
        max_rooms: int = -1,
        seed: int = 0,
        num_agents: int = 10,
        obstacle_size: float = 0.2,
        agent_size: float = 0.1,
        max_free_pos: Optional[int] = None,
        **labmaze_kwargs,
    ) -> base_map:
        map_str_batch, free_pos_str_batch = generate_labmaze_maps(
            num_maps=num_maps,
            height=height,
            width=width,
            max_rooms=max_rooms,
            seed=seed,
            **labmaze_kwargs,
        )

        super().__init__(
            map_str_batch=map_str_batch,
            free_pos_str_batch=free_pos_str_batch,
            agent_idx_batch=None,
            goal_idx_batch=None,
            num_agents=num_agents,
            random_agents=True,
            random_goals=True,
            remove_border=False,
            add_border=False,
            obstacle_size=obstacle_size,
            agent_size=agent_size,
            max_free_pos=max_free_pos,
        )
